//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint.h"

#include "JavaGpuElint.h"

// JAVA WEIRDNESS: the JNI API is not the same in C and C++!
#if defined __cplusplus
# define ENV_FNCT_1(env,function,arg_1) \
         ((env)->function(arg_1))
# define ENV_FNCT_3(env,function,arg_1,arg_2,arg_3) \
         ((env)->function(arg_1,arg_2,arg_3))
#else
# define ENV_FNCT_1(env,function,arg_1) \
         ((*(env))->function((env),arg_1))
# define ENV_FNCT_3(env,function,arg_1,arg_2,arg_3) \
         ((*(env))->function((env),arg_1,arg_2,arg_3))
#endif

jlong
JNICALL
Java_JavaGpuElint__1new(JNIEnv *env,
                        jclass cls,
                        jlong sharedNtv)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return (jlong)(intptr_t)GpuElint_new((const GpuElint *)(intptr_t)sharedNtv);
}

void
JNICALL
Java_JavaGpuElint__1delete(JNIEnv *env,
                           jclass cls,
                           jlong ntv)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
GpuElint_delete((GpuElint *)(intptr_t)ntv);
}

jboolean
JNICALL
Java_JavaGpuElint__1properties(JNIEnv *env,
                               jclass cls,
                               jlong ntv,
                               jobject out_properties)
{
(void)cls; // avoid ``unused parameter'' warning
const GpuElintProperties *p=GpuElint_properties((GpuElint *)(intptr_t)ntv);
if(!p) return JNI_FALSE;
jclass pCls=ENV_FNCT_1(env,GetObjectClass,out_properties);
jfieldID f;
f=ENV_FNCT_3(env,GetFieldID,pCls,"DTOA_SKIP_COUNT","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->DTOA_SKIP_COUNT);
f=ENV_FNCT_3(env,GetFieldID,pCls,"MAX_MATCH_COUNT","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->MAX_MATCH_COUNT);
f=ENV_FNCT_3(env,GetFieldID,pCls,"glVersionMajor","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->glVersionMajor);
f=ENV_FNCT_3(env,GetFieldID,pCls,"glVersionMinor","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->glVersionMinor);
f=ENV_FNCT_3(env,GetFieldID,pCls,"glslVersion","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->glslVersion);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxWorkGroupCountX","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxWorkGroupCount[0]);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxWorkGroupCountY","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxWorkGroupCount[1]);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxWorkGroupCountZ","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxWorkGroupCount[2]);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxWorkGroupSizeX","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxWorkGroupSize[0]);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxWorkGroupSizeY","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxWorkGroupSize[1]);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxWorkGroupSizeZ","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxWorkGroupSize[2]);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxWorkGroupInvocations","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxWorkGroupInvocations);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxUniformBlocks","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxUniformBlocks);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxTextureImageUnits","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxTextureImageUnits);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxAtomicCounterBuffers","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxAtomicCounterBuffers);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxAtomicCounters","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxAtomicCounters);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxSharedMemorySize","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxSharedMemorySize);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxUniformComponents","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxUniformComponents);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxImageUniforms","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxImageUniforms);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxCombinedUniformComponents","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxCombinedUniformComponents);
f=ENV_FNCT_3(env,GetFieldID,pCls,"maxShaderStorageBlocks","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->maxShaderStorageBlocks);
f=ENV_FNCT_3(env,GetFieldID,pCls,"warpSize","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->warpSize);
f=ENV_FNCT_3(env,GetFieldID,pCls,"warpsPerStreamingMultiprocessor","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->warpsPerStreamingMultiprocessor);
f=ENV_FNCT_3(env,GetFieldID,pCls,"streamingMultiprocessorCount","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_properties,f,p->streamingMultiprocessorCount);
f=ENV_FNCT_3(env,GetFieldID,pCls,"hasBallot","Z");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetBooleanField,out_properties,f,p->hasBallot);
f=ENV_FNCT_3(env,GetFieldID,pCls,"hasShuffle","Z");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetBooleanField,out_properties,f,p->hasShuffle);
return JNI_TRUE;
}

jboolean
JNICALL
Java_JavaGpuElint__1flush(JNIEnv *env,
                          jclass cls,
                          jlong ntv,
                          jboolean wait)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_flush((GpuElint *)(intptr_t)ntv,
                      wait ? true : false)
       ? JNI_TRUE : JNI_FALSE;
}

jint
JNICALL
Java_JavaGpuElint__1gpuAlloc(JNIEnv *env,
                             jclass cls,
                             jlong ntv,
                             jlong capacity)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_gpuAlloc((GpuElint *)(intptr_t)ntv,
                         (intptr_t)capacity);
}

jboolean
JNICALL
Java_JavaGpuElint__1gpuFree(JNIEnv *env,
                            jclass cls,
                            jlong ntv,
                            jint gpuBuffer)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_gpuFree((GpuElint *)(intptr_t)ntv,
                        gpuBuffer)
       ? JNI_TRUE : JNI_FALSE;
}

jboolean
JNICALL
Java_JavaGpuElint__1hostToGpu(JNIEnv *env,
                              jclass cls,
                              jlong ntv,
                              jint gpuDst,
                              jobject hostSrc,
                              jlong dstOffset,
                              jlong srcOffset,
                              jlong size)
{
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_hostToGpu((GpuElint *)(intptr_t)ntv,
                          gpuDst,
                          ENV_FNCT_1(env,GetDirectBufferAddress,hostSrc),
                          (intptr_t)dstOffset,
                          (intptr_t)srcOffset,
                          (intptr_t)size)
       ? JNI_TRUE : JNI_FALSE;
}

jboolean
JNICALL
Java_JavaGpuElint__1gpuToHost(JNIEnv *env,
                              jclass cls,
                              jlong ntv,
                              jobject hostDst,
                              jint gpuSrc,
                              jlong dstOffset,
                              jlong srcOffset,
                              jlong size)
{
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_gpuToHost((GpuElint *)(intptr_t)ntv,
                          ENV_FNCT_1(env,GetDirectBufferAddress,hostDst),
                          gpuSrc,
                          (intptr_t)dstOffset,
                          (intptr_t)srcOffset,
                          (intptr_t)size)
       ? JNI_TRUE : JNI_FALSE;
}

jboolean
JNICALL
Java_JavaGpuElint__1gpuToGpu(JNIEnv *env,
                             jclass cls,
                             jlong ntv,
                             jint gpuDst,
                             jint gpuSrc,
                             jlong dstOffset,
                             jlong srcOffset,
                             jlong size)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_gpuToGpu((GpuElint *)(intptr_t)ntv,
                         gpuDst,
                         gpuSrc,
                         (intptr_t)dstOffset,
                         (intptr_t)srcOffset,
                         (intptr_t)size)
       ? JNI_TRUE : JNI_FALSE;
}

jint
JNICALL
Java_JavaGpuElint__1createMarker(JNIEnv *env,
                                 jclass cls,
                                 jlong ntv)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_createMarker((GpuElint *)(intptr_t)ntv);
}

jboolean
JNICALL
Java_JavaGpuElint__1destroyMarker(JNIEnv *env,
                                  jclass cls,
                                  jlong ntv,
                                  jint marker)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_destroyMarker((GpuElint *)(intptr_t)ntv,
                              marker)
       ? JNI_TRUE : JNI_FALSE;
}

jboolean
JNICALL
Java_JavaGpuElint__1setMarker(JNIEnv *env,
                              jclass cls,
                              jlong ntv,
                              jint marker)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_setMarker((GpuElint *)(intptr_t)ntv,
                          marker)
       ? JNI_TRUE : JNI_FALSE;
}

jint
JNICALL
Java_JavaGpuElint__1testMarker(JNIEnv *env,
                               jclass cls,
                               jlong ntv,
                               jint marker)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_testMarker((GpuElint *)(intptr_t)ntv,
                           marker);
}

jfloat
JNICALL
Java_JavaGpuElint__1markerInterval(JNIEnv *env,
                                   jclass cls,
                                   jlong ntv,
                                   jint firstMarker,
                                   jint secondMarker)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_markerInterval((GpuElint *)(intptr_t)ntv,
                               firstMarker,
                               secondMarker);
}

jboolean
JNICALL
Java_JavaGpuElint__1zeroFill(JNIEnv *env,
                             jclass cls,
                             jlong ntv,
                             jint gpuDst,
                             jint begin,
                             jint end)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_zeroFill((GpuElint *)(intptr_t)ntv,
                         gpuDst,
                         begin,
                         end)
       ? JNI_TRUE : JNI_FALSE;
}

jboolean
JNICALL
Java_JavaGpuElint__1countFill(JNIEnv *env,
                              jclass cls,
                              jlong ntv,
                              jint gpuDst,
                              jint begin,
                              jint end,
                              jint countStart,
                              jint countIncrement)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_countFill((GpuElint *)(intptr_t)ntv,
                          gpuDst,
                          begin,
                          end,
                          countStart,
                          countIncrement)
       ? JNI_TRUE : JNI_FALSE;
}

jint
JNICALL
Java_JavaGpuElint__1modes1D(JNIEnv *env,
                            jclass cls,
                            jlong ntv,
                            jobject hostDst_modes,
                            jint gpuSrc_counts,
                            jfloat cutoffRatio,
                            jint binCount,
                            jboolean cyclic)
{
(void)cls; // avoid ``unused parameter'' warning
void *modes=ENV_FNCT_1(env,GetDirectBufferAddress,hostDst_modes);
jlong capacity=ENV_FNCT_1(env,GetDirectBufferCapacity,hostDst_modes);
return GpuElint_modes1D((GpuElint *)(intptr_t)ntv,
                        (HistoMode1D *)modes,
                        (int)(capacity/sizeof(HistoMode1D)),
                        gpuSrc_counts,
                        cutoffRatio,
                        binCount,
                        cyclic==JNI_TRUE);
}

jint
JNICALL
Java_JavaGpuElint__1modes2D(JNIEnv *env,
                            jclass cls,
                            jlong ntv,
                            jobject hostDst_modes,
                            jint gpuSrc_counts,
                            jfloat cutoffRatio,
                            jint binCountA,
                            jboolean cyclicA,
                            jint binCountB,
                            jboolean cyclicB)
{
(void)cls; // avoid ``unused parameter'' warning
void *modes=ENV_FNCT_1(env,GetDirectBufferAddress,hostDst_modes);
jlong capacity=ENV_FNCT_1(env,GetDirectBufferCapacity,hostDst_modes);
return GpuElint_modes2D((GpuElint *)(intptr_t)ntv,
                        (HistoMode2D *)modes,
                        (int)(capacity/sizeof(HistoMode2D)),
                        gpuSrc_counts,
                        cutoffRatio,
                        binCountA,
                        cyclicA==JNI_TRUE,
                        binCountB,
                        cyclicB==JNI_TRUE);
}

jint
JNICALL
Java_JavaGpuElint__1modes3D(JNIEnv *env,
                            jclass cls,
                            jlong ntv,
                            jobject hostDst_modes,
                            jint gpuSrc_counts,
                            jfloat cutoffRatio,
                            jint binCountA,
                            jboolean cyclicA,
                            jint binCountB,
                            jboolean cyclicB,
                            jint binCountC,
                            jboolean cyclicC)
{
(void)cls; // avoid ``unused parameter'' warning
void *modes=ENV_FNCT_1(env,GetDirectBufferAddress,hostDst_modes);
jlong capacity=ENV_FNCT_1(env,GetDirectBufferCapacity,hostDst_modes);
return GpuElint_modes3D((GpuElint *)(intptr_t)ntv,
                        (HistoMode3D *)modes,
                        (int)(capacity/sizeof(HistoMode3D)),
                        gpuSrc_counts,
                        cutoffRatio,
                        binCountA,
                        cyclicA==JNI_TRUE,
                        binCountB,
                        cyclicB==JNI_TRUE,
                        binCountC,
                        cyclicC==JNI_TRUE);
}

jint
JNICALL
Java_JavaGpuElint__1modes4D(JNIEnv *env,
                            jclass cls,
                            jlong ntv,
                            jobject hostDst_modes,
                            jint gpuSrc_counts,
                            jfloat cutoffRatio,
                            jint binCountA,
                            jboolean cyclicA,
                            jint binCountB,
                            jboolean cyclicB,
                            jint binCountC,
                            jboolean cyclicC,
                            jint binCountD,
                            jboolean cyclicD)
{
(void)cls; // avoid ``unused parameter'' warning
void *modes=ENV_FNCT_1(env,GetDirectBufferAddress,hostDst_modes);
jlong capacity=ENV_FNCT_1(env,GetDirectBufferCapacity,hostDst_modes);
return GpuElint_modes4D((GpuElint *)(intptr_t)ntv,
                        (HistoMode4D *)modes,
                        (int)(capacity/sizeof(HistoMode4D)),
                        gpuSrc_counts,
                        cutoffRatio,
                        binCountA,
                        cyclicA==JNI_TRUE,
                        binCountB,
                        cyclicB==JNI_TRUE,
                        binCountC,
                        cyclicC==JNI_TRUE,
                        binCountD,
                        cyclicD==JNI_TRUE);
}

jfloat
JNICALL
Java_JavaGpuElint__1estimateMajorPri(JNIEnv *env,
                                     jclass cls,
                                     jlong ntv,
                                     jobject hostSrc_counts,
                                     jobject hostSrc_sums,
                                     jint binCount,
                                     jfloat minDtoa,
                                     jfloat dtoaRatio,
                                     jfloat exclusivePriLowerLimit,
                                     jfloat minDtoaSum)
{
(void)cls; // avoid ``unused parameter'' warning
void *counts=ENV_FNCT_1(env,GetDirectBufferAddress,hostSrc_counts);
void *sums=ENV_FNCT_1(env,GetDirectBufferAddress,hostSrc_sums);
return GpuElint_estimateMajorPri((GpuElint *)(intptr_t)ntv,
                                 (const int *)counts,
                                 (const float *)sums,
                                 binCount,
                                 minDtoa,
                                 dtoaRatio,
                                 exclusivePriLowerLimit,
                                 minDtoaSum);
}

jboolean
JNICALL
Java_JavaGpuElint__1minMax(JNIEnv *env,
                           jclass cls,
                           jlong ntv,
                           jint gpuSrc_indices,
                           jint pulseCount,
                           jint gpuSrc_values,
                           jobject out_result)
{
(void)cls; // avoid ``unused parameter'' warning
float minValue,maxValue;
int minId,maxId;
if(!GpuElint_ID_minMax((GpuElint *)(intptr_t)ntv,
                       gpuSrc_indices,
                       pulseCount,
                       gpuSrc_values,
                       &minValue,
                       &minId,
                       &maxValue,
                       &maxId))
  {
  return JNI_FALSE;
  }
jclass rCls=ENV_FNCT_1(env,GetObjectClass,out_result);
jfieldID f;
f=ENV_FNCT_3(env,GetFieldID,rCls,"minValue","F");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetFloatField,out_result,f,minValue);
f=ENV_FNCT_3(env,GetFieldID,rCls,"minId","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_result,f,minId);
f=ENV_FNCT_3(env,GetFieldID,rCls,"maxValue","F");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetFloatField,out_result,f,maxValue);
f=ENV_FNCT_3(env,GetFieldID,rCls,"maxId","I");
ON_ERROR(!f,"invalid field",{ return JNI_FALSE; });
ENV_FNCT_3(env,SetIntField,out_result,f,maxId);
return JNI_TRUE;
}

jboolean
JNICALL
Java_JavaGpuElint__1histogram1D(JNIEnv *env,
                                jclass cls,
                                jlong ntv,
                                jint gpuSrc_indices,
                                jint pulseCount,
                                jint gpuDst_counts,
                                jint gpuSrc_values,
                                jfloat minValue,
                                jfloat maxValue,
                                jint binCount)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_histogram1D((GpuElint *)(intptr_t)ntv,
                               gpuSrc_indices,
                               pulseCount,
                               gpuDst_counts,
                               gpuSrc_values,
                               minValue,
                               maxValue,
                               binCount)
       ? JNI_TRUE : JNI_FALSE;
}

jboolean
JNICALL
Java_JavaGpuElint__1histogram2D(JNIEnv *env,
                                jclass cls,
                                jlong ntv,
                                jint gpuSrc_indices,
                                jint pulseCount,
                                jint gpuDst_counts,
                                jint gpuSrc_valuesA,
                                jfloat minValueA,
                                jfloat maxValueA,
                                jint binCountA,
                                jint gpuSrc_valuesB,
                                jfloat minValueB,
                                jfloat maxValueB,
                                jint binCountB)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_histogram2D((GpuElint *)(intptr_t)ntv,
                               gpuSrc_indices,
                               pulseCount,
                               gpuDst_counts,
                               gpuSrc_valuesA,
                               minValueA,
                               maxValueA,
                               binCountA,
                               gpuSrc_valuesB,
                               minValueB,
                               maxValueB,
                               binCountB)
       ? JNI_TRUE : JNI_FALSE;
}

jboolean
JNICALL
Java_JavaGpuElint__1histogram3D(JNIEnv *env,
                                jclass cls,
                                jlong ntv,
                                jint gpuSrc_indices,
                                jint pulseCount,
                                jint gpuDst_counts,
                                jint gpuSrc_valuesA,
                                jfloat minValueA,
                                jfloat maxValueA,
                                jint binCountA,
                                jint gpuSrc_valuesB,
                                jfloat minValueB,
                                jfloat maxValueB,
                                jint binCountB,
                                jint gpuSrc_valuesC,
                                jfloat minValueC,
                                jfloat maxValueC,
                                jint binCountC)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_histogram3D((GpuElint *)(intptr_t)ntv,
                               gpuSrc_indices,
                               pulseCount,
                               gpuDst_counts,
                               gpuSrc_valuesA,
                               minValueA,
                               maxValueA,
                               binCountA,
                               gpuSrc_valuesB,
                               minValueB,
                               maxValueB,
                               binCountB,
                               gpuSrc_valuesC,
                               minValueC,
                               maxValueC,
                               binCountC)
       ? JNI_TRUE : JNI_FALSE;
}

jboolean
JNICALL
Java_JavaGpuElint__1histogram4D(JNIEnv *env,
                                jclass cls,
                                jlong ntv,
                                jint gpuSrc_indices,
                                jint pulseCount,
                                jint gpuDst_counts,
                                jint gpuSrc_valuesA,
                                jfloat minValueA,
                                jfloat maxValueA,
                                jint binCountA,
                                jint gpuSrc_valuesB,
                                jfloat minValueB,
                                jfloat maxValueB,
                                jint binCountB,
                                jint gpuSrc_valuesC,
                                jfloat minValueC,
                                jfloat maxValueC,
                                jint binCountC,
                                jint gpuSrc_valuesD,
                                jfloat minValueD,
                                jfloat maxValueD,
                                jint binCountD)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_histogram4D((GpuElint *)(intptr_t)ntv,
                               gpuSrc_indices,
                               pulseCount,
                               gpuDst_counts,
                               gpuSrc_valuesA,
                               minValueA,
                               maxValueA,
                               binCountA,
                               gpuSrc_valuesB,
                               minValueB,
                               maxValueB,
                               binCountB,
                               gpuSrc_valuesC,
                               minValueC,
                               maxValueC,
                               binCountC,
                               gpuSrc_valuesD,
                               minValueD,
                               maxValueD,
                               binCountD)
       ? JNI_TRUE : JNI_FALSE;
}

jint
JNICALL
Java_JavaGpuElint__1filter1D(JNIEnv *env,
                             jclass cls,
                             jlong ntv,
                             jint gpuSrc_indices,
                             jint pulseCount,
                             jint gpuDst_indices,
                             jint gpuSrc_values,
                             jfloat minValue,
                             jfloat maxValue)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_filter1D((GpuElint *)(intptr_t)ntv,
                            gpuSrc_indices,
                            pulseCount,
                            gpuDst_indices,
                            gpuSrc_values,
                            minValue,
                            maxValue);
}

jint
JNICALL
Java_JavaGpuElint__1filter2D(JNIEnv *env,
                             jclass cls,
                             jlong ntv,
                             jint gpuSrc_indices,
                             jint pulseCount,
                             jint gpuDst_indices,
                             jint gpuSrc_valuesA,
                             jfloat minValueA,
                             jfloat maxValueA,
                             jint gpuSrc_valuesB,
                             jfloat minValueB,
                             jfloat maxValueB)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_filter2D((GpuElint *)(intptr_t)ntv,
                            gpuSrc_indices,
                            pulseCount,
                            gpuDst_indices,
                            gpuSrc_valuesA,
                            minValueA,
                            maxValueA,
                            gpuSrc_valuesB,
                            minValueB,
                            maxValueB);
}

jint
JNICALL
Java_JavaGpuElint__1filter3D(JNIEnv *env,
                             jclass cls,
                             jlong ntv,
                             jint gpuSrc_indices,
                             jint pulseCount,
                             jint gpuDst_indices,
                             jint gpuSrc_valuesA,
                             jfloat minValueA,
                             jfloat maxValueA,
                             jint gpuSrc_valuesB,
                             jfloat minValueB,
                             jfloat maxValueB,
                             jint gpuSrc_valuesC,
                             jfloat minValueC,
                             jfloat maxValueC)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_filter3D((GpuElint *)(intptr_t)ntv,
                            gpuSrc_indices,
                            pulseCount,
                            gpuDst_indices,
                            gpuSrc_valuesA,
                            minValueA,
                            maxValueA,
                            gpuSrc_valuesB,
                            minValueB,
                            maxValueB,
                            gpuSrc_valuesC,
                            minValueC,
                            maxValueC);
}

jint
JNICALL
Java_JavaGpuElint__1filter4D(JNIEnv *env,
                             jclass cls,
                             jlong ntv,
                             jint gpuSrc_indices,
                             jint pulseCount,
                             jint gpuDst_indices,
                             jint gpuSrc_valuesA,
                             jfloat minValueA,
                             jfloat maxValueA,
                             jint gpuSrc_valuesB,
                             jfloat minValueB,
                             jfloat maxValueB,
                             jint gpuSrc_valuesC,
                             jfloat minValueC,
                             jfloat maxValueC,
                             jint gpuSrc_valuesD,
                             jfloat minValueD,
                             jfloat maxValueD)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_filter4D((GpuElint *)(intptr_t)ntv,
                            gpuSrc_indices,
                            pulseCount,
                            gpuDst_indices,
                            gpuSrc_valuesA,
                            minValueA,
                            maxValueA,
                            gpuSrc_valuesB,
                            minValueB,
                            maxValueB,
                            gpuSrc_valuesC,
                            minValueC,
                            maxValueC,
                            gpuSrc_valuesD,
                            minValueD,
                            maxValueD);
}

jboolean
JNICALL
Java_JavaGpuElint__1dtoaHistogram(JNIEnv *env,
                                  jclass cls,
                                  jlong ntv,
                                  jint gpuSrc_indices,
                                  jint pulseCount,
                                  jint gpuSrc_toa64,
                                  jobject hostDst_counts,
                                  jobject hostDst_sums,
                                  jint binCount,
                                  jfloat minDtoa,
                                  jfloat dtoaRatio)
{
(void)cls; // avoid ``unused parameter'' warning
void *counts=ENV_FNCT_1(env,GetDirectBufferAddress,hostDst_counts);
void *sums=ENV_FNCT_1(env,GetDirectBufferAddress,hostDst_sums);
return GpuElint_ID_dtoaHistogram((GpuElint *)(intptr_t)ntv,
                                 gpuSrc_indices,
                                 pulseCount,
                                 gpuSrc_toa64,
                                 (int *)counts,
                                 (float *)sums,
                                 binCount,
                                 minDtoa,
                                 dtoaRatio)
       ? JNI_TRUE : JNI_FALSE;
}

jfloat
JNICALL
Java_JavaGpuElint__1markPulses(JNIEnv *env,
                               jclass cls,
                               jlong ntv,
                               jint gpuSrc_indices,
                               jint pulseCount,
                               jint gpuSrc_toa64,
                               jint gpu_marks,
                               jint mark,
                               jfloat estimatedPri,
                               jfloat priRatioThreshold,
                               jint minMatchCount)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_markPulses((GpuElint *)(intptr_t)ntv,
                              gpuSrc_indices,
                              pulseCount,
                              gpuSrc_toa64,
                              gpu_marks,
                              mark,
                              estimatedPri,
                              priRatioThreshold,
                              minMatchCount);
}

jint
JNICALL
Java_JavaGpuElint__1filterUnmarked(JNIEnv *env,
                                   jclass cls,
                                   jlong ntv,
                                   jint gpuSrc_indices,
                                   jint pulseCount,
                                   jint gpuDst_indices,
                                   jint gpuSrc_marks)
{
(void)env; // avoid ``unused parameter'' warning
(void)cls; // avoid ``unused parameter'' warning
return GpuElint_ID_filterUnmarked((GpuElint *)(intptr_t)ntv,
                                  gpuSrc_indices,
                                  pulseCount,
                                  gpuDst_indices,
                                  gpuSrc_marks);
}

//----------------------------------------------------------------------------
