//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_context.h"

#if defined _WIN32

// FIXME: explain why a so ugly solution is needed

#define UGLY_WGL_CONTEXT_POOL_SIZE 16 // adjust as needed

struct _UglyWglContextPool
{
CRITICAL_SECTION sync;
bool underDestruction;
HGLRC wglCtx[UGLY_WGL_CONTEXT_POOL_SIZE];
bool inUse[UGLY_WGL_CONTEXT_POOL_SIZE];
};

static
void
_UglyWglContextPool_destroy(struct _UglyWglContextPool *pool)
{
// first was captured, not created
for(int i=1;i<UGLY_WGL_CONTEXT_POOL_SIZE;++i)
  {
  if(pool->wglCtx[i])
    {
    wglDeleteContext(pool->wglCtx[i]);
    }
  }
DeleteCriticalSection(&pool->sync);
Memory_free(pool);
}

static
struct _UglyWglContextPool *
_UglyWglContextPool_init(HDC hdc,
                         HGLRC rootCtx)
{
struct _UglyWglContextPool *pool;
pool=Memory_initAlloc(1,struct _UglyWglContextPool);
InitializeCriticalSection(&pool->sync);
HGLRC (APIENTRY *wglCreateContextAttribs)(HDC,HGLRC,const int *);
const char * suffixes[]={ "", "ARB", "EXT", (const char *)0 };
CoreGl_extensionFunctionWithSuffix(&wglCreateContextAttribs, // may fail
                                   "wglCreateContextAttribs",
                                   suffixes);
pool->wglCtx[0]=rootCtx; // first is captured, not created
for(int i=1;i<UGLY_WGL_CONTEXT_POOL_SIZE;++i)
  {
  if(wglCreateContextAttribs) // modern means to create a shared context
    {
    pool->wglCtx[i]=(*wglCreateContextAttribs)(hdc,rootCtx,(const int *)0);
    ON_ERROR(!pool->wglCtx[i],"wglCreateContextAttribs() failure",
                              { goto undo; });
    }
  else // fallback to legacy shared context creation
    {
    pool->wglCtx[i]=wglCreateContext(hdc);
    ON_ERROR(!pool->wglCtx[i],"wglCreateContext() failure",{ goto undo; });
    BOOL success=wglShareLists(rootCtx,pool->wglCtx[i]);
    ON_ERROR(!success,"wglShareLists() failure",{ goto undo; });
    }
  }
return pool;
goto undo; // avoid ``unused label'' warning
undo:
_UglyWglContextPool_destroy(pool);
return (struct _UglyWglContextPool *)0;
}

static
HGLRC
_UglyWglContextPool_useContext(struct _UglyWglContextPool *pool)
{
HGLRC ctx=(HGLRC)0;
EnterCriticalSection(&pool->sync);
if(!pool->underDestruction)
  {
  // first was captured, not created
  for(int i=1;i<UGLY_WGL_CONTEXT_POOL_SIZE;++i)
    {
    if(!pool->inUse[i])
      {
      ctx=pool->wglCtx[i];
      pool->inUse[i]=true;
      break;
      }
    }
  }
LeaveCriticalSection(&pool->sync);
return ctx;
}

static
void
_UglyWglContextPool_releaseContext(struct _UglyWglContextPool *pool,
                                   HGLRC ctx)
{
EnterCriticalSection(&pool->sync);
int inUseCount=0;
for(int i=0;i<UGLY_WGL_CONTEXT_POOL_SIZE;++i)
  {
  if(pool->wglCtx[i]==ctx)
    {
    pool->inUse[i]=false;
    if(i&&(ctx==wglGetCurrentContext())) // first was captured, not created
      {
      wglMakeCurrent((HDC)0,(HGLRC)0);
      }
    }
  if(pool->inUse[i])
    {
    ++inUseCount;
    }
  }
if(!inUseCount)
  {
  pool->underDestruction=true;
  }
LeaveCriticalSection(&pool->sync);
if(!inUseCount)
  {
  // FIXME: a newly created shared context in another thread may be trying to
  //        enter the critical section while it is under destruction.
  //        This is very unlikely although, because any new shared context
  //        refers to a preceding one so the pool can't be under destruction.
  _UglyWglContextPool_destroy(pool);
  }
}

bool // success
GpuElintContext_init(GpuElintContext *c,
                     const GpuElintContext *shared)
{
c->thread=GetCurrentThreadId();
if(shared)
  {
  c->hdc=shared->hdc;
  c->wglCtxPool=shared->wglCtxPool;
  c->ctx=_UglyWglContextPool_useContext(c->wglCtxPool);
  ON_ERROR(!c->ctx,"WGL shared context pool exhauted",{ return false; });
  BOOL success=wglMakeCurrent(c->hdc,c->ctx);
  ON_ERROR(!success,"wglMakeCurrent() failure",
                    {
                    _UglyWglContextPool_releaseContext(c->wglCtxPool,c->ctx);
                    return false;
                    });
  }
else
  {
  c->hdc=wglGetCurrentDC();
  ON_ERROR(!c->hdc,"wglGetCurrentDC() failure",{ return false; });
  c->ctx=wglGetCurrentContext();
  ON_ERROR(!c->ctx,"wglGetCurrentContext() failure",{ return false; });
  if(!CoreGl_init()) { return false; }
  c->wglCtxPool=_UglyWglContextPool_init(c->hdc,c->ctx);
  if(!c->wglCtxPool) { return false; }
  }
c->initDone=true;
return true;
}

void
GpuElintContext_destroy(GpuElintContext *c)
{
if(!c->initDone) return;
_UglyWglContextPool_releaseContext(c->wglCtxPool,c->ctx);
}

bool // thread/context mismatch detected
GpuElintContext_mismatch(const GpuElintContext *c)
{
return !((GetCurrentThreadId()==c->thread)&&
         (wglGetCurrentContext()==c->ctx));
}

#else // not _WIN32

bool // success
GpuElintContext_init(GpuElintContext *c,
                     const GpuElintContext *shared)
{
c->thread=pthread_self();
if(shared)
  {
  c->shared=true;
  c->dpy=shared->dpy;
  c->drawable=shared->drawable;
  c->isDirect=shared->isDirect;
  c->fbConfig=shared->fbConfig;
  c->renderType=shared->renderType;
  // FIXME: in OpenTK the reported renderType is 0!
  c->renderType=GLX_RGBA_TYPE;
  c->screen=shared->screen;
  c->ctx=glXCreateNewContext(c->dpy,c->fbConfig,c->renderType,
                             shared->ctx,c->isDirect);
  ON_ERROR(!c->ctx,"glXCreateNewContext() failure",{ return false; });
  Bool success=glXMakeCurrent(c->dpy,c->drawable,c->ctx);
  ON_ERROR(!success,"glXMakeCurrent() failure",
           { glXDestroyContext(c->dpy,c->ctx); return false; });
  }
else
  {
  c->shared=false;
  c->dpy=glXGetCurrentDisplay();
  ON_ERROR(!c->dpy,"glXGetCurrentDisplay() failure",{ return false; });
  c->drawable=glXGetCurrentDrawable();
  ON_ERROR(!c->drawable,"glXGetCurrentDrawable() failure",{ return false; });
  c->ctx=glXGetCurrentContext();
  ON_ERROR(!c->ctx,"glXGetCurrentContext() failure",{ return false; });
  c->isDirect=glXIsDirect(c->dpy,c->ctx);
  int success,fbConfigId;
  success=glXQueryContext(c->dpy,c->ctx,GLX_FBCONFIG_ID,&fbConfigId);
  ON_ERROR(success!=Success,"glXQueryContext() failure",{ return false; });
  success=glXQueryContext(c->dpy,c->ctx,GLX_RENDER_TYPE,&c->renderType);
  ON_ERROR(success!=Success,"glXQueryContext() failure",{ return false; });
  success=glXQueryContext(c->dpy,c->ctx,GLX_SCREEN,&c->screen);
  ON_ERROR(success!=Success,"glXQueryContext() failure",{ return false; });
  int attrList[]={ GLX_FBCONFIG_ID, fbConfigId, None };
  int fbConfigCount;
  GLXFBConfig *fbConfigList=glXChooseFBConfig(c->dpy,c->screen,
                                              attrList,&fbConfigCount);
  ON_ERROR(!fbConfigList,"glXChooseFBConfig() failure",{ return false; });
  c->fbConfig=*fbConfigList;
  XFree(fbConfigList);
  if(!CoreGl_init())
    { return false; }
  }
c->initDone=true;
return true;
}

void
GpuElintContext_destroy(GpuElintContext *c)
{
if(!c->initDone) return;
if(c->shared)
  {
  glXDestroyContext(c->dpy,c->ctx);
  }
}

bool // thread/context mismatch detected
GpuElintContext_mismatch(const GpuElintContext *c)
{
return !(pthread_equal(pthread_self(),c->thread)&&
         (glXGetCurrentContext()==c->ctx));
}

#endif // not _WIN32

//----------------------------------------------------------------------------
