//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#ifndef GPUELINT_INTERNALS_H
#define GPUELINT_INTERNALS_H 1

#include "gpuElint.h"
#include "gpuElint_context.h"

struct _GpuElint
{
int64_t initTime;
GpuElintContext context;
GpuElintProperties properties;
int gpuBackend;
intptr_t gpuBackendCapacity;
struct _GpuServiceCountFill *countFill_service;
struct _GpuServiceModes1D *modes1D_service;
struct _GpuServiceModes2D *modes2D_service;
struct _GpuServiceModes3D *modes3D_service;
struct _GpuServiceModes4D *modes4D_service;
//
struct _GpuService_ID_Compact *ID_compact_service;
struct _GpuService_ID_MinMax *ID_minMax_service;
struct _GpuService_ID_Histogram1D *ID_histogram1D_service;
struct _GpuService_ID_Histogram2D *ID_histogram2D_service;
struct _GpuService_ID_Histogram3D *ID_histogram3D_service;
struct _GpuService_ID_Histogram4D *ID_histogram4D_service;
struct _GpuService_ID_Filter1D *ID_filter1D_service;
struct _GpuService_ID_Filter2D *ID_filter2D_service;
struct _GpuService_ID_Filter3D *ID_filter3D_service;
struct _GpuService_ID_Filter4D *ID_filter4D_service;
struct _GpuService_ID_DtoaHistogram *ID_dtoaHistogram_service;
struct _GpuService_ID_MarkPulses *ID_markPulses_service;
struct _GpuService_ID_FilterUnmarked *ID_filterUnmarked_service;
//
struct _GpuService_BM_MinMax *BM_minMax_service;
struct _GpuService_BM_Histogram1D *BM_histogram1D_service;
struct _GpuService_BM_Histogram2D *BM_histogram2D_service;
struct _GpuService_BM_Histogram3D *BM_histogram3D_service;
struct _GpuService_BM_Histogram4D *BM_histogram4D_service;
struct _GpuService_BM_Filter1D *BM_filter1D_service;
struct _GpuService_BM_Filter2D *BM_filter2D_service;
struct _GpuService_BM_Filter3D *BM_filter3D_service;
struct _GpuService_BM_Filter4D *BM_filter4D_service;
struct _GpuService_BM_DtoaHistogram *BM_dtoaHistogram_service;
struct _GpuService_BM_MarkPulses *BM_markPulses_service;
struct _GpuService_BM_FilterUnmarked *BM_filterUnmarked_service;
};

#define GPUELINT_API_INVOCATION_CHECK(ge,action)                  \
        do                                                        \
          {                                                       \
          ON_ERROR(!ge,"null GpuElint object",action);            \
          bool __mismatch=GpuElintContext_mismatch(&ge->context); \
          ON_ERROR(__mismatch,"thread/context mismatch",action);  \
          glGetError(); /* clear any previous error */            \
          } while(0)

int // identifier of the backend GPU-buffer (0 on failure)
GpuElint_gpuBackend(GpuElint *ge,
                    intptr_t capacity);

int // maximal size supported by a 1D work-group
GpuElint_maxWorkGroupSize(const GpuElint *ge);

int // maximal power-of-two size supported by a 1D work-group
GpuElint_maxPowerOfTwoWorkGroupSize(const GpuElint *ge);

int // a generaly suitable work-group size
GpuElint_chooseWorkGroupSize(const GpuElint *ge,
                             bool powerOfTwo);

int // a generaly suitable work-group count
GpuElint_chooseWorkGroupCount(const GpuElint *ge);

PRINTF_FORMAT(5,6)
void
GpuElint_appendText(GpuElint *ge,
                    char **inout_storage,
                    int *inout_size,
                    int *inout_capacity,
                    const char *format,
                    ...);

void
GpuElint_appendCodePreamble(GpuElint *ge,
                            char **inout_storage,
                            int *inout_size,
                            int *inout_capacity,
                            int groupSize,
                            int groupCount);

GLuint // built GPU program (0 on failure)
GpuElint_buildProgram(GpuElint *ge,
                      const char *sourceCode);

int // identifier of the backend GPU-buffer for compaction (0 on failure)
GpuElint_ID_compactionBackend(GpuElint *ge,
                              int pulseCount);

int // number of remaining pulses (-1 on failure)
GpuElint_ID_compact(GpuElint *ge,
                    int gpu_backend,
                    int pulseCount,
                    int gpuDst_indices);

#endif // GPUELINT_INTERNALS_H

//----------------------------------------------------------------------------
