//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

#define UINT_REAL 16.0 // adjust as needed (watch out for overflow!)

typedef struct _GpuService_BM_DtoaHistogram
{
GLuint program;
int groupSize;
int groupCount;
} GpuService_BM_DtoaHistogram;

static
bool //success
buildProgram(GpuElint *ge,
             GpuService_BM_DtoaHistogram *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false)&~31; // multiple of 32
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " #define UINT_REAL       %f               \n"
  " #define DTOA_SKIP_COUNT %d               \n"
  "                                          \n"
  "   VOLATILE_BUFFER(0, uint   results[]; ) \n"
  "  READ_ONLY_BUFFER(1, uint   bitmask[]; ) \n"
  "  READ_ONLY_BUFFER(2, double toa64[];   ) \n"
  "                                          \n"
  " UNIFORM_PARAMETER(0, ivec2 )             \n"
  " const int pulseCount=UNI_0.x,            \n"
  "           binCount  =UNI_0.y;            \n"
  "                                          \n"
  " UNIFORM_PARAMETER(1, vec2  )             \n"
  " const float minDtoa  =UNI_1.x,           \n"
  "             dtoaRatio=UNI_1.y;           \n"
  "                                          \n",
  UINT_REAL,
  GPUELINT_DTOA_SKIP_COUNT);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " int maskId=-1;                                \n"
  " uint mask=0;                                  \n"
  "                                               \n"
  " uint                                          \n"
  " loadMask(int id)                              \n"
  " {                                             \n"
  " int mId=id>>5;                                \n"
  " if(maskId!=mId) { mask=bitmask[maskId=mId]; } \n"
  " return mask;                                  \n"
  " }                                             \n"
  "                                               \n");
#if 0
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " void                                                              \n"
    " main(void) // BM_dtoahistogram()                                  \n"
    " {                                                                 \n"
    " const float log_minDtoa=log(minDtoa);                             \n"
    " const float log_dtoaRatio=log(dtoaRatio);                         \n"
    " const float div_log_dtoaRatio=1.0/log_dtoaRatio;                  \n"
    " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)                  \n"
    "   {                                                               \n"
    "   if((loadMask(id)&(1<<(id&31)))!=0)                              \n"
    "     {                                                             \n"
    "     const double toa=toa64[id];                                   \n"
    "     for(int next=id,skipCount=0;++next<pulseCount;)               \n"
    "       {                                                           \n"
    "       if((loadMask(next)&(1<<(next&31)))!=0)                      \n"
    "         {                                                         \n"
    "         const float log_dtoa=log(abs(float(toa64[next]-toa)));    \n"
    "         const float pos=(log_dtoa-log_minDtoa)*div_log_dtoaRatio; \n"
    "         const int bin=int(floor(pos));                            \n"
    "         if((bin>=0)&&(bin<binCount))                              \n"
    "           {                                                       \n"
    "           const uint uiFract=uint(round(UINT_REAL*fract(pos)));   \n"
    "           atomicAdd(results[bin],1);                              \n"
    "           atomicAdd(results[binCount+bin],uiFract);               \n"
    "           }                                                       \n"
    "         if(++skipCount==DTOA_SKIP_COUNT) { break; }               \n"
    "         }                                                         \n"
    "       }                                                           \n"
    "     }                                                             \n"
    "   }                                                               \n"
    " }                                                                 \n");
#else // caching bitmask into shared memory seems helpful here
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " shared uint sh_mask[GROUP_SIZE];                                  \n"
    "                                                                   \n"
    " void                                                              \n"
    " main(void) // BM_dtoahistogram()                                  \n"
    " {                                                                 \n"
    " const float log_minDtoa=log(minDtoa);                             \n"
    " const float log_dtoaRatio=log(dtoaRatio);                         \n"
    " const float div_log_dtoaRatio=1.0/log_dtoaRatio;                  \n"
    " for(int id=groupId*(GROUP_SIZE<<5)+localId;                       \n"
    "     id<pulseCount;id+=GRID_SIZE<<5)                               \n"
    "   {                                                               \n"
    "   const int grpBegin=id-localId;                                  \n"
    "   const int grpEnd=min(grpBegin+(GROUP_SIZE<<5),pulseCount);      \n"
    "   barrier(); // no early overwrite of sh_mask[]                   \n"
    "   sh_mask[localId]=bitmask[(grpBegin>>5)+localId];                \n"
    "   memoryBarrierShared();                                          \n"
    "   barrier(); // wait for sh_mask[]                                \n"
    "   double firstToa=0.0lf;                                          \n"
    "   int skipCount=0,first=id,i=id;                                  \n"
    "   while(first<grpEnd)                                             \n"
    "     {                                                             \n"
    "     bool reset=false;                                             \n"
    "     const uint mask=i<grpEnd ? sh_mask[(i-grpBegin)>>5] :         \n"
    "                     i<pulseCount ? loadMask(i) : 0;               \n"
    "     if((mask&(1<<(i&31)))!=0)                                     \n"
    "       {                                                           \n"
    "       const double toa=toa64[i];                                  \n"
    "       if(skipCount==0) { firstToa=toa; }                          \n"
    "       else                                                        \n"
    "         {                                                         \n"
    "         const float log_dtoa=log(abs(float(toa-firstToa)));       \n"
    "         const float pos=(log_dtoa-log_minDtoa)*div_log_dtoaRatio; \n"
    "         const int bin=int(floor(pos));                            \n"
    "         if((bin>=0)&&(bin<binCount))                              \n"
    "           {                                                       \n"
    "           const uint uiFract=uint(round(UINT_REAL*fract(pos)));   \n"
    "           atomicAdd(results[bin],1);                              \n"
    "           atomicAdd(results[binCount+bin],uiFract);               \n"
    "           }                                                       \n"
    "         }                                                         \n"
    "       if(skipCount++==DTOA_SKIP_COUNT) { reset=true; }            \n"
    "       }                                                           \n"
    "     else                                                          \n"
    "       {                                                           \n"
    "       if((skipCount==0)||(i>=pulseCount)) { reset=true; }         \n"
    "       }                                                           \n"
    "     if(reset) { skipCount=0; i=(first+=GROUP_SIZE); }             \n"
    "     else { ++i; }                                                 \n"
    "     }                                                             \n"
    "   }                                                               \n"
    " }                                                                 \n");
#endif
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_BM_DtoaHistogram_init(GpuElint *ge)
{
GpuService_BM_DtoaHistogram *s=Memory_alloc(1,GpuService_BM_DtoaHistogram);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->BM_dtoaHistogram_service=s;
return true;
}

void
GpuService_BM_DtoaHistogram_destroy(GpuElint *ge)
{
GpuService_BM_DtoaHistogram *s=ge->BM_dtoaHistogram_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

bool // success
GpuElint_BM_dtoaHistogram(GpuElint *ge,
                          int gpuSrc_bitmask,
                          int pulseCount,
                          int gpuSrc_toa64,
                          int *hostDst_counts,
                          float *hostDst_sums,
                          int binCount,
                          float minDtoa,
                          float dtoaRatio)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
const GpuService_BM_DtoaHistogram *s=ge->BM_dtoaHistogram_service;
const int histoSize=binCount*sizeof(int32_t);
int gpu_results=GpuElint_gpuBackend(ge,2*histoSize);
ON_ERROR(!gpu_results,"GpuElint_gpuBackend() failure",{ return false; });
bool success=GpuElint_zeroFill(ge,gpu_results,0,2*histoSize);
ON_ERROR(!success,"GpuElint_zeroFill() failure",{ return false; });
glUseProgram(s->program);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpu_results);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_bitmask);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_toa64);
ON_GL_ERROR({ return false; });
glUniform2i(0,pulseCount,binCount);
ON_GL_ERROR({ return false; });
glUniform2f(1,minDtoa,dtoaRatio);
ON_GL_ERROR({ return false; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return false; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return false; });
success=GpuElint_gpuToHost(ge,hostDst_counts,gpu_results,0,0,histoSize);
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return false; });
success=GpuElint_gpuToHost(ge,hostDst_sums,gpu_results,0,histoSize,histoSize);
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return false; });
const double log_minDtoa=log(minDtoa);
const double log_dtoaRatio=log(dtoaRatio);
const double fractFactor=1.0/UINT_REAL;
float *sums=hostDst_sums;
const unsigned int *uiSums=(const unsigned int *)hostDst_sums;
int *counts=hostDst_counts;
for(int bin=0;bin<binCount;++bin)
  {
  const int count=counts[bin];
  const unsigned int uiSum=uiSums[bin];
  const double sum=bin+(uiSum*fractFactor)/ONE_IF_ZERO(count);
  sums[bin]=(float)(count*exp((log_dtoaRatio*sum)+log_minDtoa));
  }
return true;
}

//----------------------------------------------------------------------------
