//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuService_BM_Filter1D
{
GLuint program;
int groupSize;
int groupCount;
} GpuService_BM_Filter1D;

static
bool // success
buildProgram(GpuElint *ge,
             GpuService_BM_Filter1D *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
const int warpSize=ge->properties.warpSize;
const bool hasBallot=(warpSize==32)&&ge->properties.hasBallot;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false)&~31; // multiple of 32
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " WRITE_ONLY_BUFFER(0, uint  dstBitmask[]; ) \n"
  "  READ_ONLY_BUFFER(1, uint  srcBitmask[]; ) \n"
  "  READ_ONLY_BUFFER(2, float values[];     ) \n"
  "                                            \n"
  " UNIFORM_PARAMETER(0, int  )                \n"
  " const int pulseCount=UNI_0.x;              \n"
  "                                            \n"
  " UNIFORM_PARAMETER(1, vec2 )                \n"
  " const float minValue=UNI_1.x,              \n"
  "             maxValue=UNI_1.y;              \n"
  "                                            \n");
if(hasBallot)
  {
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " uint                        \n"
    " ballot(bool bit)            \n"
    " {                           \n"
    " return ballotThreadNV(bit); \n"
    " }                           \n"
    "                             \n");
  }
else
  {
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " shared uint sh_ballot[GROUP_SIZE>>5];             \n"
    "                                                   \n"
    " // FIXME: extremely slow emulation!               \n"
    " uint                                              \n"
    " ballot(bool bit)                                  \n"
    " {                                                 \n"
    " const int maskId=localId>>5;                      \n"
    " const uint bitShift=1<<(localId&31);              \n"
    " barrier(); // no early overwrite of sh_ballot[];  \n"
    " if(bitShift==1) { sh_ballot[maskId]=0; }          \n"
    " memoryBarrierShared();                            \n"
    " barrier(); // wait for sh_ballot[];               \n"
    " if(bit) { atomicOr(sh_ballot[maskId],bitShift); } \n"
    " memoryBarrierShared();                            \n"
    " barrier(); // wait for sh_ballot[];               \n"
    " return sh_ballot[maskId];                         \n"
    " }                                                 \n"
    "                                                   \n");
  }
#if 0
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " void                                                \n"
    " main(void) // BM_filter1D()                         \n"
    " {                                                   \n"
    " const bool wrap=minValue>maxValue;                  \n"
    " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)    \n"
    "   {                                                 \n"
    "   const int maskId=id>>5;                           \n"
    "   bool keep=(srcBitmask[maskId]&(1<<(id&31)))!=0;   \n"
    "   if(keep)                                          \n"
    "     {                                               \n"
    "     const float value=values[id];                   \n"
    "     const bool c1=value>=minValue;                  \n"
    "     const bool c2=value<=maxValue;                  \n"
    "     keep=wrap ? (c1||c2) : (c1&&c2);                \n"
    "     }                                               \n"
    "   const uint keepMask=ballot(keep);                 \n"
    "   if(inWarpId==0) { dstBitmask[maskId]=keepMask; }  \n"
    "   }                                                 \n"
    " }                                                   \n");
#else // caching bitmask into shared memory seems helpful here
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " shared uint sh_mask[GROUP_SIZE];                             \n"
    "                                                              \n"
    " void                                                         \n"
    " main(void) // BM_filter1D()                                  \n"
    " {                                                            \n"
    " const uint testMask=1<<(localId&31);                         \n"
    " const bool wrap=minValue>maxValue;                           \n"
    " for(int id=groupId*(GROUP_SIZE<<5)+localId;                  \n"
    "     id<pulseCount;id+=GRID_SIZE<<5)                          \n"
    "   {                                                          \n"
    "   const int grpBegin=id-localId;                             \n"
    "   const int grpEnd=min(grpBegin+(GROUP_SIZE<<5),pulseCount); \n"
    "   barrier(); // no early overwrite of sh_mask[]              \n"
    "   sh_mask[localId]=srcBitmask[(grpBegin>>5)+localId];        \n"
    "   memoryBarrierShared();                                     \n"
    "   barrier(); // wait for sh_mask[]                           \n"
    "   for(int i=id;i<grpEnd;i+=GROUP_SIZE)                       \n"
    "     {                                                        \n"
    "     const uint mask=sh_mask[(i-grpBegin)>>5];                \n"
    "     bool keep=(mask&testMask)!=0;                            \n"
    "     if(keep)                                                 \n"
    "       {                                                      \n"
    "       const float value=values[i];                           \n"
    "       const bool c1=value>=minValue;                         \n"
    "       const bool c2=value<=maxValue;                         \n"
    "       keep=wrap ? (c1||c2) : (c1&&c2);                       \n"
    "       }                                                      \n"
    "     const uint keepMask=ballot(keep);                        \n"
    "     if(testMask==1) { dstBitmask[i>>5]=keepMask; }           \n"
    "     }                                                        \n"
    "   }                                                          \n"
    " }                                                            \n");
#endif
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_BM_Filter1D_init(GpuElint *ge)
{
GpuService_BM_Filter1D *s=Memory_alloc(1,GpuService_BM_Filter1D);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->BM_filter1D_service=s;
return true;
}

void
GpuService_BM_Filter1D_destroy(GpuElint *ge)
{
GpuService_BM_Filter1D *s=ge->BM_filter1D_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

bool // success
GpuElint_BM_filter1D(GpuElint *ge,
                     int gpuSrc_bitmask,
                     int pulseCount,
                     int gpuDst_bitmask,
                     int gpuSrc_values,
                     float minValue,
                     float maxValue)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
const GpuService_BM_Filter1D *s=ge->BM_filter1D_service;
glUseProgram(s->program);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpuDst_bitmask);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_bitmask);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_values);
ON_GL_ERROR({ return false; });
glUniform1i(0,pulseCount);
ON_GL_ERROR({ return false; });
glUniform2f(1,minValue,maxValue);
ON_GL_ERROR({ return false; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return false; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return false; });
return true;
}

//----------------------------------------------------------------------------
