//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuService_BM_Histogram2D
{
GLuint program;
int groupSize;
int groupCount;
} GpuService_BM_Histogram2D;

static
bool // success
buildProgram(GpuElint *ge,
             GpuService_BM_Histogram2D *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false)&~31; // multiple of 32
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  "   VOLATILE_BUFFER(0, int   counts[];  ) \n"
  "  READ_ONLY_BUFFER(1, uint  bitmask[]; ) \n"
  "  READ_ONLY_BUFFER(2, float valuesA[]; ) \n"
  "  READ_ONLY_BUFFER(3, float valuesB[]; ) \n"
  "                                         \n"
  " UNIFORM_PARAMETER(0, ivec3 )            \n"
  " const int pulseCount=UNI_0.x;           \n"
  " const ivec2 binCount=UNI_0.yz;          \n"
  "                                         \n"
  " UNIFORM_PARAMETER(1, vec4  )            \n"
  " const vec2 factor   =UNI_1.xy,          \n"
  "            increment=UNI_1.zw;          \n"
  "                                         \n");
#if 0
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " void                                                           \n"
    " main(void) // BM_histogram2D()                                 \n"
    " {                                                              \n"
    " const ivec2 firstBin=ivec2(0);                                 \n"
    " const ivec2 lastBin=binCount-ivec2(1);                         \n"
    " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)               \n"
    "   {                                                            \n"
    "   if((bitmask[id>>5]&(1<<(id&31)))!=0)                         \n"
    "     {                                                          \n"
    "     const vec2 value=vec2(valuesA[id],                         \n"
    "                           valuesB[id]);                        \n"
    "     const vec2 pos=fma(value,factor,increment);                \n"
    "     const ivec2 bin=clamp(ivec2(floor(pos)),firstBin,lastBin); \n"
    "     const int binId=bin.y*binCount.x+bin.x;                    \n"
    "     atomicAdd(counts[binId],1);                                \n"
    "     }                                                          \n"
    "   }                                                            \n"
    " }                                                              \n");
#else // caching bitmask into shared memory seems helpful here
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " shared uint sh_mask[GROUP_SIZE];                                 \n"
    "                                                                  \n"
    " void                                                             \n"
    " main(void) // BM_histogram2D()                                   \n"
    " {                                                                \n"
    " const ivec2 firstBin=ivec2(0);                                   \n"
    " const ivec2 lastBin=binCount-ivec2(1);                           \n"
    " const uint testMask=1<<(localId&31);                             \n"
    " for(int id=groupId*(GROUP_SIZE<<5)+localId;                      \n"
    "     id<pulseCount;id+=GRID_SIZE<<5)                              \n"
    "   {                                                              \n"
    "   const int grpBegin=id-localId;                                 \n"
    "   const int grpEnd=min(grpBegin+(GROUP_SIZE<<5),pulseCount);     \n"
    "   barrier(); // no early overwrite of sh_mask[]                  \n"
    "   sh_mask[localId]=bitmask[(grpBegin>>5)+localId];               \n"
    "   memoryBarrierShared();                                         \n"
    "   barrier(); // wait for sh_mask[]                               \n"
    "   for(int i=id;i<grpEnd;i+=GROUP_SIZE)                           \n"
    "     {                                                            \n"
    "     const uint mask=sh_mask[(i-grpBegin)>>5];                    \n"
    "     if((mask&testMask)!=0)                                       \n"
    "       {                                                          \n"
    "       const vec2 value=vec2(valuesA[i],                          \n"
    "                             valuesB[i]);                         \n"
    "       const vec2 pos=fma(value,factor,increment);                \n"
    "       const ivec2 bin=clamp(ivec2(floor(pos)),firstBin,lastBin); \n"
    "       const int binId=bin.y*binCount.x+bin.x;                    \n"
    "       atomicAdd(counts[binId],1);                                \n"
    "       }                                                          \n"
    "     }                                                            \n"
    "   }                                                              \n"
    " }                                                                \n");
#endif
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_BM_Histogram2D_init(GpuElint *ge)
{
GpuService_BM_Histogram2D *s=Memory_alloc(1,GpuService_BM_Histogram2D);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->BM_histogram2D_service=s;
return true;
}

void
GpuService_BM_Histogram2D_destroy(GpuElint *ge)
{
GpuService_BM_Histogram2D *s=ge->BM_histogram2D_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

bool // success
GpuElint_BM_histogram2D(GpuElint *ge,
                        int gpuSrc_bitmask,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_valuesA,
                        float minValueA,
                        float maxValueA,
                        int binCountA,
                        int gpuSrc_valuesB,
                        float minValueB,
                        float maxValueB,
                        int binCountB)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
const GpuService_BM_Histogram2D *s=ge->BM_histogram2D_service;
// binCount*(v-minValue)/(maxValue-minValue)
const float factorA=binCountA/(maxValueA-minValueA);
const float factorB=binCountB/(maxValueB-minValueB);
const float incrementA=-minValueA*factorA;
const float incrementB=-minValueB*factorB;
bool success=GpuElint_zeroFill(ge,gpuDst_counts,0,binCountA*binCountB);
ON_ERROR(!success,"GpuElint_zeroFill() failure",{ return false; });
glUseProgram(s->program);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpuDst_counts);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_bitmask);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_valuesA);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,3,gpuSrc_valuesB);
ON_GL_ERROR({ return false; });
glUniform3i(0,pulseCount,binCountA,binCountB);
ON_GL_ERROR({ return false; });
glUniform4f(1,factorA,factorB,incrementA,incrementB);
ON_GL_ERROR({ return false; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return false; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return false; });
return true;
}

//----------------------------------------------------------------------------
