//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _Result
{
int priCount;
float meanPri;
} Result;

typedef struct _GpuService_BM_MarkPulses
{
GLuint program;
int groupSize;
int groupCount;
int gpu_results;
Result *host_results;
} GpuService_BM_MarkPulses;

static
bool // success
buildProgram(GpuElint *ge,
             GpuService_BM_MarkPulses *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
const int warpSize=ge->properties.warpSize;
const bool hasShuffle=ge->properties.hasShuffle;
// a power-of-two group size is MANDATORY for reduction
s->groupSize=GpuElint_chooseWorkGroupSize(ge,true);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
const int sharedSize=hasShuffle ? (s->groupSize+warpSize-1)/warpSize
                                : s->groupSize;
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " #define MAX_MATCH_COUNT %d                       \n"
  "                                                  \n"
  " struct Result { int priCount; float meanPri; };  \n"
  "                                                  \n"
  " shared Result sh_results[%d];                    \n"
  "                                                  \n"
  " // shared memory is faster than local arrays     \n"
  " shared int matchMem[GROUP_SIZE*MAX_MATCH_COUNT]; \n"
  "                                                  \n"
  " #define MATCH_MEM(m)                           \\\n"
  "         (matchMem[(m)*GROUP_SIZE+localId])       \n"
  "                                                  \n"
  " WRITE_ONLY_BUFFER(0, Result results[]; )         \n"
  " WRITE_ONLY_BUFFER(1, int    marks[];   )         \n"
  "  READ_ONLY_BUFFER(2, uint   bitmask[]; )         \n"
  "  READ_ONLY_BUFFER(3, double toa64[];   )         \n"
  "                                                  \n"
  " UNIFORM_PARAMETER(0, ivec3 )                     \n"
  " const int pulseCount   =UNI_0.x,                 \n"
  "           minMatchCount=UNI_0.y,                 \n"
  "           mark         =UNI_0.z;                 \n"
  "                                                  \n"
  " UNIFORM_PARAMETER(1, vec2  )                     \n"
  " const float estimatedPri     =UNI_1.x,           \n"
  "             priRatioThreshold=UNI_1.y;           \n"
  "                                                  \n",
  GPUELINT_MAX_MATCH_COUNT,
  sharedSize);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " int maskId=-1;                                         \n"
  " uint mask=0;                                           \n"
  "                                                        \n"
  " uint                                                   \n"
  " loadMask(int id)                                       \n"
  " {                                                      \n"
  " int mId=id>>5;                                         \n"
  " if(maskId!=mId) { mask=bitmask[maskId=mId]; }          \n"
  " return mask;                                           \n"
  " }                                                      \n"
  "                                                        \n"
  " void                                                   \n"
  " mergeResults(inout Result r,                           \n"
  "              const Result tmp)                         \n"
  " {                                                      \n"
  " if(tmp.priCount==0) { return; }                        \n"
  " if(r.priCount==0) { r=tmp; return; }                   \n"
  " const double meanPri=double(r.meanPri)*r.priCount+     \n"
  "                      double(tmp.meanPri)*tmp.priCount; \n"
  " r.priCount+=tmp.priCount;                              \n"
  " r.meanPri=float(meanPri/r.priCount);                   \n"
  " }                                                      \n"
  "                                                        \n");
#if 1
  // this uncached version is simpler than the caching alternative
  // and finally gives better performances
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " void                                                            \n"
    " main(void) // BM_markPulses()                                   \n"
    " {                                                               \n"
    " int priCount=0;                                                 \n"
    " double priSum=0.0lf;                                            \n"
    " float expectedPri=estimatedPri;                                 \n"
    " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)                \n"
    "   {                                                             \n"
    "   if((loadMask(id)&(1<<(id&31)))!=0)                            \n"
    "     {                                                           \n"
    "     float localPriSum=0.0;                                      \n"
    "     double toa=toa64[id];                                       \n"
    "     double expectedToa=toa+expectedPri;                         \n"
    "     int matchCount=0,last=id;                                   \n"
    "     double lastToa=toa;                                         \n"
    "     float lastDtoa=+1.0/0.0;                                    \n"
    "     for(int next=id;++next<=pulseCount;)                        \n"
    "       {                                                         \n"
    "       bool useLast=true;                                        \n"
    "       if(next<pulseCount)                                       \n"
    "         {                                                       \n"
    "         if((loadMask(next)&(1<<(next&31)))==0)                  \n"
    "           { useLast=false; }                                    \n"
    "         else                                                    \n"
    "           {                                                     \n"
    "           const double nextToa=toa64[next];                     \n"
    "           const float nextDtoa=abs(float(expectedToa-nextToa)); \n"
    "           if(nextDtoa<lastDtoa)                                 \n"
    "             {                                                   \n"
    "             last=next;                                          \n"
    "             lastToa=nextToa;                                    \n"
    "             lastDtoa=nextDtoa;                                  \n"
    "             useLast=false;                                      \n"
    "             }                                                   \n"
    "           }                                                     \n"
    "         }                                                       \n");
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    "       if(useLast)                                               \n"
    "         {                                                       \n"
    "         const float pri=float(lastToa-toa);                     \n"
    "         if((expectedPri>pri*priRatioThreshold)||                \n"
    "            (pri>expectedPri*priRatioThreshold))                 \n"
    "           { break; }                                            \n"
    "         localPriSum+=pri;                                       \n"
    "         MATCH_MEM(matchCount)=last;                             \n"
    "         if(++matchCount==minMatchCount) { break; }              \n"
    "         next=last;                                              \n"
    "         lastDtoa=+1.0/0.0;                                      \n"
    "         toa=0.5lf*(lastToa+expectedToa);                        \n"
    "         expectedToa=toa+expectedPri;                            \n"
    "         }                                                       \n"
    "       }                                                         \n"
    "     if(matchCount==minMatchCount)                               \n"
    "       {                                                         \n"
    "       priCount+=minMatchCount;                                  \n"
    "       priSum+=localPriSum;                                      \n"
    "       expectedPri=float(priSum/priCount);                       \n"
    "       marks[id]=mark;                                           \n"
    "       for(int m=0;m<minMatchCount;++m)                          \n"
    "         { marks[MATCH_MEM(m)]=mark; }                           \n"
    "       }                                                         \n"
    "     }                                                           \n"
    "   }                                                             \n"
    " Result r={ priCount, priCount!=0 ? expectedPri : 0.0 };         \n");
#else // caching bitmask into shared memory does NOT seem helpful here
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " shared uint sh_mask[GROUP_SIZE];                                  \n"
    "                                                                   \n"
    " void                                                              \n"
    " main(void) // BM_markPulses()                                     \n"
    " {                                                                 \n"
    " int priCount=0;                                                   \n"
    " double priSum=0.0lf;                                              \n"
    " float expectedPri=estimatedPri;                                   \n"
    " for(int id=groupId*(GROUP_SIZE<<5)+localId;                       \n"
    "     id<pulseCount;id+=GRID_SIZE<<5)                               \n"
    "   {                                                               \n"
    "   const int grpBegin=id-localId;                                  \n"
    "   const int grpEnd=min(grpBegin+(GROUP_SIZE<<5),pulseCount);      \n"
    "   barrier(); // no early overwrite of sh_mask[]                   \n"
    "   sh_mask[localId]=bitmask[(grpBegin>>5)+localId];                \n"
    "   memoryBarrierShared();                                          \n"
    "   barrier(); // wait for sh_mask[]                                \n"
    "   for(int i=id;i<grpEnd;i+=GROUP_SIZE)                            \n"
    "     {                                                             \n"
    "     const uint mask=sh_mask[(i-grpBegin)>>5];                     \n"
    "     if((mask&(1<<(i&31)))!=0)                                     \n"
    "       {                                                           \n"
    "       float localPriSum=0.0;                                      \n"
    "       double toa=toa64[i];                                        \n"
    "       double expectedToa=toa+expectedPri;                         \n"
    "       int matchCount=0,last=i;                                    \n"
    "       double lastToa=toa;                                         \n"
    "       float lastDtoa=+1.0/0.0;                                    \n"
    "       for(int next=i;++next<=pulseCount;)                         \n"
    "         {                                                         \n"
    "         bool useLast=true;                                        \n"
    "         if(next<pulseCount)                                       \n"
    "           {                                                       \n"
    "           const uint nextMask=next<grpEnd                         \n"
    "                               ? sh_mask[(next-grpBegin)>>5]       \n"
    "                               : loadMask(next);                   \n"
    "           if((nextMask&(1<<(next&31)))==0)                        \n"
    "             { useLast=false; }                                    \n"
    "           else                                                    \n"
    "             {                                                     \n"
    "             const double nextToa=toa64[next];                     \n"
    "             const float nextDtoa=abs(float(expectedToa-nextToa)); \n"
    "             if(nextDtoa<lastDtoa)                                 \n"
    "               {                                                   \n"
    "               last=next;                                          \n"
    "               lastToa=nextToa;                                    \n"
    "               lastDtoa=nextDtoa;                                  \n"
    "               useLast=false;                                      \n"
    "               }                                                   \n"
    "             }                                                     \n"
    "           }                                                       \n");
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    "         if(useLast)                                               \n"
    "           {                                                       \n"
    "           const float pri=float(lastToa-toa);                     \n"
    "           if((expectedPri>pri*priRatioThreshold)||                \n"
    "              (pri>expectedPri*priRatioThreshold))                 \n"
    "             { break; }                                            \n"
    "           localPriSum+=pri;                                       \n"
    "           MATCH_MEM(matchCount)=last;                             \n"
    "           if(++matchCount==minMatchCount) { break; }              \n"
    "           next=last;                                              \n"
    "           lastDtoa=+1.0/0.0;                                      \n"
    "           toa=0.5lf*(lastToa+expectedToa);                        \n"
    "           expectedToa=toa+expectedPri;                            \n"
    "           }                                                       \n"
    "         }                                                         \n"
    "       if(matchCount==minMatchCount)                               \n"
    "         {                                                         \n"
    "         priCount+=minMatchCount;                                  \n"
    "         priSum+=localPriSum;                                      \n"
    "         expectedPri=float(priSum/priCount);                       \n"
    "         marks[i]=mark;                                            \n"
    "         for(int m=0;m<minMatchCount;++m)                          \n"
    "           { marks[MATCH_MEM(m)]=mark; }                           \n"
    "         }                                                         \n"
    "       }                                                           \n"
    "     }                                                             \n"
    "   }                                                               \n"
    " Result r={ priCount, priCount!=0 ? expectedPri : 0.0 };           \n");
#endif
if(hasShuffle)
  {
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " Result tmp;      \n"
    " int shuffleId;   \n"
    " uint activeMask; \n");
  for(int sz=s->groupSize;sz>1;sz/=warpSize)
    {
    GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
      " activeMask=ballotThreadNV(true); \n");
    for(int i=warpSize/2;i;i/=2)
      {
      if(sz>i)
        {
        GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
          " shuffleId=inWarpId+%d;                                  \n"
          " if((activeMask&(1<<shuffleId))==0)                      \n"
          "   { shuffleId=inWarpId; }; // avoid inactive threads    \n"
          " tmp.priCount=shuffleNV(r.priCount,shuffleId,WARP_SIZE); \n"
          " tmp.meanPri=shuffleNV(r.meanPri,shuffleId,WARP_SIZE);   \n"
          " mergeResults(r,tmp);                                    \n",
          i);
        }
      }
    if(sz>warpSize)
      {
      GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
        " if(inWarpId==0) { sh_results[warpId]=r; } \n"
        " memoryBarrierShared();                    \n"
        " barrier(); // wait for sh_results[]       \n"
        " if(localId>=%d) { return; }               \n"
        " r=sh_results[localId];                    \n",
        sz/warpSize);
      if(sz>warpSize*warpSize)
        {
        GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
          " barrier(); // no early overwrite of sh_results[] \n");
        }
      }
    }
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " if(localId!=0) { return; } \n");
  }
else
  {
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " sh_results[localId]=r;              \n"
    " memoryBarrierShared();              \n"
    " barrier(); // wait for sh_results[] \n");
  for(int i=s->groupSize>>1;i>1;i>>=1)
    {
    GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
      " if(localId>=%d) { return; }             \n"
      " mergeResults(r,sh_results[localId+%d]); \n"
      " // keep barriers within warps since     \n"
      " // lock-step assumption is unsafe       \n"
      " memoryBarrierShared();                  \n"
      " barrier(); // wait for sh_results[]     \n",
      i,i);
    }
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " if(localId!=0) { return; }     \n"
    " mergeResults(r,sh_results[1]); \n");
  }
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " results[groupId]=r; \n"
  " }                   \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_BM_MarkPulses_init(GpuElint *ge)
{
GpuService_BM_MarkPulses *s=Memory_alloc(1,GpuService_BM_MarkPulses);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
s->gpu_results=GpuElint_gpuAlloc(ge,s->groupCount*sizeof(Result));
ON_ERROR(!s->gpu_results,"GpuElint_gpuAlloc() failure",
         {
         glDeleteProgram(s->program);
         Memory_free(s);
         return false;
         });
s->host_results=Memory_alloc(s->groupCount,Result);
ge->BM_markPulses_service=s;
return true;
}

void
GpuService_BM_MarkPulses_destroy(GpuElint *ge)
{
GpuService_BM_MarkPulses *s=ge->BM_markPulses_service;
if(!s) return;
Memory_free(s->host_results);
bool success=GpuElint_gpuFree(ge,s->gpu_results);
ON_ERROR(!success,"GpuElint_gpuFree() failure",{});
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

float // actual pulse-repetition-interval or 0.0f (-1.0f on failure)
GpuElint_BM_markPulses(GpuElint *ge,
                       int gpuSrc_bitmask,
                       int pulseCount,
                       int gpuSrc_toa64,
                       int gpu_marks,
                       int mark,
                       float estimatedPri,
                       float priRatioThreshold,
                       int minMatchCount)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1.0f; });
const GpuService_BM_MarkPulses *s=ge->BM_markPulses_service;
ON_ERROR(minMatchCount>GPUELINT_MAX_MATCH_COUNT,
         "MAX_MATCH_COUNT exceeded",{ return -1.0f; });
const int groupCount=s->groupCount;
glUseProgram(s->program);
ON_GL_ERROR({ return -1.0f; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,s->gpu_results);
ON_GL_ERROR({ return -1.0f; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpu_marks);
ON_GL_ERROR({ return -1.0f; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_bitmask);
ON_GL_ERROR({ return -1.0f; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,3,gpuSrc_toa64);
ON_GL_ERROR({ return -1.0f; });
glUniform3i(0,pulseCount,minMatchCount,mark);
ON_GL_ERROR({ return -1.0f; });
glUniform2f(1,estimatedPri,priRatioThreshold);
ON_GL_ERROR({ return -1.0f; });
glDispatchCompute(groupCount,1,1);
ON_GL_ERROR({ return -1.0f; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return -1.0f; });
Result *host_results=s->host_results;
bool success=GpuElint_gpuToHost(ge,host_results,s->gpu_results,
                                0,0,groupCount*sizeof(Result));
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return -1.0f; });
int priCount=0;
double priSum=0.0;
for(int i=0;i<groupCount;++i)
  {
  const Result r=host_results[i];
  priCount+=r.priCount;
  priSum+=r.priCount*r.meanPri;
  }
return priCount ? (float)(priSum/priCount) : 0.0f;
}

//----------------------------------------------------------------------------
