//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _ValueId
{
float value;
int id;
} ValueId;

typedef struct _Result
{
ValueId minPair,maxPair;
} Result;

typedef struct _GpuService_BM_MinMax
{
GLuint program;
int groupSize;
int groupCount;
int gpu_results;
Result *host_results;
} GpuService_BM_MinMax;

static
bool // success
buildProgram(GpuElint *ge,
             GpuService_BM_MinMax *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
const int warpSize=ge->properties.warpSize;
const bool hasShuffle=ge->properties.hasShuffle;
// a power-of-two group size is MANDATORY for reduction
s->groupSize=GpuElint_chooseWorkGroupSize(ge,true);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
const int sharedSize=hasShuffle ? (s->groupSize+warpSize-1)/warpSize
                                : s->groupSize;
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " struct ValueId { float value; int id; };     \n"
  " struct Result  { ValueId minPair,maxPair; }; \n"
  "                                              \n"
  " shared ValueId sh_min[%d];                   \n"
  " shared ValueId sh_max[%d];                   \n"
  "                                              \n"
  " WRITE_ONLY_BUFFER(0, Result results[]; )     \n"
  "  READ_ONLY_BUFFER(1, uint   bitmask[]; )     \n"
  "  READ_ONLY_BUFFER(2, float  values[];  )     \n"
  "                                              \n"
  " UNIFORM_PARAMETER(0, int )                   \n"
  " const int pulseCount=UNI_0.x;                \n"
  "                                              \n",
  sharedSize,sharedSize);
#if 0
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " void                                               \n"
    " main(void) // BM_minMax()                          \n"
    " {                                                  \n"
    " ValueId minPair={ +1.0/0.0, -1 },                  \n"
    "         maxPair={ -1.0/0.0, -1 },                  \n"
    "         p;                                         \n"
    " for(p.id=globalId;p.id<pulseCount;p.id+=GRID_SIZE) \n"
    "   {                                                \n"
    "   if((bitmask[p.id>>5]&(1<<(p.id&31)))!=0)         \n"
    "     {                                              \n"
    "     p.value=values[p.id];                          \n"
    "     if(p.value<minPair.value) { minPair=p; }       \n"
    "     if(p.value>maxPair.value) { maxPair=p; }       \n"
    "     }                                              \n"
    "   }                                                \n");
#else // caching bitmask into shared memory seems helpful here
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " shared uint sh_mask[GROUP_SIZE];                             \n"
    "                                                              \n"
    " void                                                         \n"
    " main(void) // BM_minMax()                                    \n"
    " {                                                            \n"
    " ValueId minPair={ +1.0/0.0, -1 },                            \n"
    "         maxPair={ -1.0/0.0, -1 },                            \n"
    "         p;                                                   \n"
    " const uint testMask=1<<(localId&31);                         \n"
    " for(int id=groupId*(GROUP_SIZE<<5)+localId;                  \n"
    "     id<pulseCount;id+=GRID_SIZE<<5)                          \n"
    "   {                                                          \n"
    "   const int grpBegin=id-localId;                             \n"
    "   const int grpEnd=min(grpBegin+(GROUP_SIZE<<5),pulseCount); \n"
    "   barrier(); // no early overwrite of sh_mask[]              \n"
    "   sh_mask[localId]=bitmask[(grpBegin>>5)+localId];           \n"
    "   memoryBarrierShared();                                     \n"
    "   barrier(); // wait for sh_mask[]                           \n"
    "   for(p.id=id;p.id<grpEnd;p.id+=GROUP_SIZE)                  \n"
    "     {                                                        \n"
    "     const uint mask=sh_mask[(p.id-grpBegin)>>5];             \n"
    "     if((mask&testMask)!=0)                                   \n"
    "       {                                                      \n"
    "       p.value=values[p.id];                                  \n"
    "       if(p.value<minPair.value) { minPair=p; }               \n"
    "       if(p.value>maxPair.value) { maxPair=p; }               \n"
    "       }                                                      \n"
    "     }                                                        \n"
    "   }                                                          \n");
#endif
if(hasShuffle)
  {
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " int shuffleId;   \n"
    " uint activeMask; \n");
  for(int sz=s->groupSize;sz>1;sz/=warpSize)
    {
    GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
      " activeMask=ballotThreadNV(true); \n");
    for(int i=warpSize/2;i;i/=2)
      {
      if(sz>i)
        {
        GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
          " shuffleId=inWarpId+%d;                                \n"
          " if((activeMask&(1<<shuffleId))==0)                    \n"
          "   { shuffleId=inWarpId; }; // avoid inactive threads  \n"
          " p.value=shuffleNV(minPair.value,shuffleId,WARP_SIZE); \n"
          " p.id=shuffleNV(minPair.id,shuffleId,WARP_SIZE);       \n"
          " if(p.value<minPair.value) { minPair=p; }              \n"
          " p.value=shuffleNV(maxPair.value,shuffleId,WARP_SIZE); \n"
          " p.id=shuffleNV(maxPair.id,shuffleId,WARP_SIZE);       \n"
          " if(p.value>maxPair.value) { maxPair=p; }              \n",
          i);
        }
      }
    if(sz>warpSize)
      {
      GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
        " if(inWarpId==0)                     \n"
        "   {                                 \n"
        "   sh_min[warpId]=minPair;           \n"
        "   sh_max[warpId]=maxPair;           \n"
        "   }                                 \n"
        " memoryBarrierShared();              \n"
        " barrier(); // wait for sh_min/max[] \n"
        " if(localId>=%d) { return; } \n"
        " minPair=sh_min[localId];    \n"
        " maxPair=sh_max[localId];    \n",
        sz/warpSize);
      if(sz>warpSize*warpSize)
        {
        GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
          " barrier(); // no early overwrite of sh_min/max[] \n");
        }
      }
    }
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " if(localId!=0) { return; } \n");
  }
else
  {
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " sh_min[localId]=minPair;            \n"
    " sh_max[localId]=maxPair;            \n"
    " memoryBarrierShared();              \n"
    " barrier(); // wait for sh_min/max[] \n");
  for(int i=s->groupSize>>1;i>1;i>>=1)
    {
    GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
      " if(localId>=%d) { return; }                              \n"
      " p=sh_min[localId+%d];                                    \n"
      " if(p.value<minPair.value) { sh_min[localId]=minPair=p; } \n"
      " p=sh_max[localId+%d];                                    \n"
      " if(p.value>maxPair.value) { sh_max[localId]=maxPair=p; } \n"
      " // keep barriers within warps since                      \n"
      " // lock-step assumption is unsafe                        \n"
      " memoryBarrierShared();                                   \n"
      " barrier(); // wait for sh_min/max[]                      \n",
      i,i,i);
    }
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " if(localId!=0) { return; }               \n"
    " p=sh_min[1];                             \n"
    " if(p.value<minPair.value) { minPair=p; } \n"
    " p=sh_max[1];                             \n"
    " if(p.value>maxPair.value) { maxPair=p; } \n");
  }
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " results[groupId].minPair=minPair; \n"
  " results[groupId].maxPair=maxPair; \n"
  " }                                 \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_BM_MinMax_init(GpuElint *ge)
{
GpuService_BM_MinMax *s=Memory_alloc(1,GpuService_BM_MinMax);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
s->gpu_results=GpuElint_gpuAlloc(ge,s->groupCount*sizeof(Result));
ON_ERROR(!s->gpu_results,"GpuElint_gpuAlloc() failure",
         {
         glDeleteProgram(s->program);
         Memory_free(s);
         return false;
         });
s->host_results=Memory_alloc(s->groupCount,Result);
ge->BM_minMax_service=s;
return true;
}

void
GpuService_BM_MinMax_destroy(GpuElint *ge)
{
GpuService_BM_MinMax *s=ge->BM_minMax_service;
if(!s) return;
Memory_free(s->host_results);
bool success=GpuElint_gpuFree(ge,s->gpu_results);
ON_ERROR(!success,"GpuElint_gpuFree() failure",{});
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

bool // success
GpuElint_BM_minMax(GpuElint *ge,
                   int gpuSrc_bitmask,
                   int pulseCount,
                   int gpuSrc_values,
                   float *out_minValue,
                   int *out_minId,
                   float *out_maxValue,
                   int *out_maxId)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
GpuService_BM_MinMax *s=ge->BM_minMax_service;
const int groupCount=s->groupCount;
glUseProgram(s->program);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,s->gpu_results);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_bitmask);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_values);
ON_GL_ERROR({ return false; });
glUniform1i(0,pulseCount);
ON_GL_ERROR({ return false; });
glDispatchCompute(groupCount,1,1);
ON_GL_ERROR({ return false; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return false; });
Result *host_results=s->host_results;
bool success=GpuElint_gpuToHost(ge,host_results,s->gpu_results,
                                0,0,groupCount*sizeof(Result));
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return false; });
Result result={ { FLT_MAX, -1 }, { -FLT_MAX, -1 } };
for(int i=0;i<groupCount;++i)
  {
  if(host_results[i].minPair.value<result.minPair.value)
    { result.minPair=host_results[i].minPair; }
  if(host_results[i].maxPair.value>result.maxPair.value)
    { result.maxPair=host_results[i].maxPair; }
  }
*out_minValue=result.minPair.value;
*out_minId=result.minPair.id;
*out_maxValue=result.maxPair.value;
*out_maxId=result.maxPair.id;
return true;
}

//----------------------------------------------------------------------------
