//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

#define UINT_REAL 16.0 // adjust as needed (watch out for overflow!)

typedef struct _GpuService_ID_DtoaHistogram
{
GLuint program;
int groupSize;
int groupCount;
} GpuService_ID_DtoaHistogram;

static
bool //success
buildProgram(GpuElint *ge,
             GpuService_ID_DtoaHistogram *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
const int warpSize=ge->properties.warpSize;
const int maxSharedMemorySize=ge->properties.maxSharedMemorySize;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
int sharedToaCapacity;
for(;;)
  {
  sharedToaCapacity=1;
  while(sharedToaCapacity<s->groupSize+GPUELINT_DTOA_SKIP_COUNT)
    { sharedToaCapacity<<=1; }
  if(sharedToaCapacity*(int)sizeof(real64_t)>maxSharedMemorySize)
    { s->groupSize-=warpSize; }
  else
    { break; }
  }
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " #define UINT_REAL       %f               \n"
  " #define DTOA_SKIP_COUNT %d               \n"
  " #define SHARED_CAPACITY %d               \n"
  "                                          \n"
  " shared double sh_toa[SHARED_CAPACITY];   \n"
  "                                          \n"
  " #define SH_TOA(i)                      \\\n"
  "         (sh_toa[(sharedOffset+(i))&    \\\n"
  "                 (SHARED_CAPACITY-1)])    \n"
  "                                          \n"
  "   VOLATILE_BUFFER(0, uint   results[]; ) \n"
  "  READ_ONLY_BUFFER(1, int    indices[]; ) \n"
  "  READ_ONLY_BUFFER(2, double toa64[];   ) \n"
  "                                          \n"
  " UNIFORM_PARAMETER(0, ivec3 )             \n"
  " const int pulseCount=UNI_0.x,            \n"
  "           perGroup  =UNI_0.y,            \n"
  "           binCount  =UNI_0.z;            \n"
  "                                          \n"
  " UNIFORM_PARAMETER(1, vec2  )             \n"
  " const float minDtoa  =UNI_1.x,           \n"
  "             dtoaRatio=UNI_1.y;           \n"
  "                                          \n",
  UINT_REAL,
  GPUELINT_DTOA_SKIP_COUNT,
  sharedToaCapacity);
#if 0
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " void                                                          \n"
    " main(void) // ID_dtoaHistogram()                              \n"
    " {                                                             \n"
    " const float log_minDtoa=log(minDtoa);                         \n"
    " const float log_dtoaRatio=log(dtoaRatio);                     \n"
    " const float div_log_dtoaRatio=1.0/log_dtoaRatio;              \n"
    " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)              \n"
    "   {                                                           \n"
    "   const double toa=toa64[indices[id]];                        \n"
    "   for(int next=min(pulseCount-1,id+DTOA_SKIP_COUNT);          \n"
    "       next>id;--next)                                         \n"
    "     {                                                         \n"
    "     const double nextToa=toa64[indices[next]];                \n"
    "     const float log_dtoa=log(abs(float(nextToa-toa)));        \n"
    "     const float pos=(log_dtoa-log_minDtoa)*div_log_dtoaRatio; \n"
    "     const int bin=int(floor(pos));                            \n"
    "     if((bin>=0)&&(bin<binCount))                              \n"
    "       {                                                       \n"
    "       const uint uiFract=uint(round(UINT_REAL*fract(pos)));   \n"
    "       atomicAdd(results[bin],1);                              \n"
    "       atomicAdd(results[binCount+bin],uiFract);               \n"
    "       }                                                       \n"
    "     }                                                         \n"
    "   }                                                           \n"
    " }                                                             \n");
#else // caching TOA into shared memory seems helpful here
  GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
    " void                                                          \n"
    " main(void) // ID_dtoaHistogram()                              \n"
    " {                                                             \n"
    " const float log_minDtoa=log(minDtoa);                         \n"
    " const float log_dtoaRatio=log(dtoaRatio);                     \n"
    " const float div_log_dtoaRatio=1.0/log_dtoaRatio;              \n"
    " const int grpBegin=groupId*perGroup;                          \n"
    " const int grpEnd=min(grpBegin+perGroup,pulseCount);           \n"
    " for(int start=grpBegin+localId,sharedOffset=0;                \n"
    "     start<grpEnd;                                             \n"
    "     start+=GROUP_SIZE,sharedOffset+=GROUP_SIZE)               \n"
    "   {                                                           \n"
    "   const int firstSharedPulse=start-localId;                   \n"
    "   const int sharedCount=min(pulseCount-firstSharedPulse,      \n"
    "                             GROUP_SIZE+DTOA_SKIP_COUNT);      \n"
    "   barrier(); // no early overwrite of sh_toa[]                \n"
    "   for(int i=localId+(sharedOffset==0 ? 0 : DTOA_SKIP_COUNT);  \n"
    "       i<sharedCount;i+=GROUP_SIZE)                            \n"
    "     {                                                         \n"
    "     const int index=indices[firstSharedPulse+i];              \n"
    "     SH_TOA(i)=toa64[index];                                   \n"
    "     }                                                         \n"
    "   memoryBarrierShared();                                      \n"
    "   barrier(); // wait for sh_toa[]                             \n"
    "   const double toa=SH_TOA(localId);                           \n"
    "   for(int next=min(sharedCount-1,localId+DTOA_SKIP_COUNT);    \n"
    "       next>localId;--next)                                    \n"
    "     {                                                         \n"
    "     const double nextToa=SH_TOA(next);                        \n"
    "     const float log_dtoa=log(abs(float(nextToa-toa)));        \n"
    "     const float pos=(log_dtoa-log_minDtoa)*div_log_dtoaRatio; \n"
    "     const int bin=int(floor(pos));                            \n"
    "     if((bin>=0)&&(bin<binCount))                              \n"
    "       {                                                       \n"
    "       const uint uiFract=uint(round(UINT_REAL*fract(pos)));   \n"
    "       atomicAdd(results[bin],1);                              \n"
    "       atomicAdd(results[binCount+bin],uiFract);               \n"
    "       }                                                       \n"
    "     }                                                         \n"
    "   }                                                           \n"
    " }                                                             \n");
#endif
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_ID_DtoaHistogram_init(GpuElint *ge)
{
GpuService_ID_DtoaHistogram *s=Memory_alloc(1,GpuService_ID_DtoaHistogram);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->ID_dtoaHistogram_service=s;
return true;
}

void
GpuService_ID_DtoaHistogram_destroy(GpuElint *ge)
{
GpuService_ID_DtoaHistogram *s=ge->ID_dtoaHistogram_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

bool // success
GpuElint_ID_dtoaHistogram(GpuElint *ge,
                          int gpuSrc_indices,
                          int pulseCount,
                          int gpuSrc_toa64,
                          int *hostDst_counts,
                          float *hostDst_sums,
                          int binCount,
                          float minDtoa,
                          float dtoaRatio)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
const GpuService_ID_DtoaHistogram *s=ge->ID_dtoaHistogram_service;
const int groupSize=s->groupSize;
const int groupCount=s->groupCount;
int perGroup=0;
while(perGroup*groupCount<pulseCount) { perGroup+=groupSize; }
const int histoSize=binCount*sizeof(int32_t);
int gpu_results=GpuElint_gpuBackend(ge,2*histoSize);
ON_ERROR(!gpu_results,"GpuElint_gpuBackend() failure",{ return false; });
bool success=GpuElint_zeroFill(ge,gpu_results,0,2*histoSize);
ON_ERROR(!success,"GpuElint_zeroFill() failure",{ return false; });
glUseProgram(s->program);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpu_results);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_indices);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_toa64);
ON_GL_ERROR({ return false; });
glUniform3i(0,pulseCount,perGroup,binCount);
ON_GL_ERROR({ return false; });
glUniform2f(1,minDtoa,dtoaRatio);
ON_GL_ERROR({ return false; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return false; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return false; });
success=GpuElint_gpuToHost(ge,hostDst_counts,gpu_results,0,0,histoSize);
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return false; });
success=GpuElint_gpuToHost(ge,hostDst_sums,gpu_results,0,histoSize,histoSize);
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return false; });
const double log_minDtoa=log(minDtoa);
const double log_dtoaRatio=log(dtoaRatio);
const double fractFactor=1.0/UINT_REAL;
float *sums=hostDst_sums;
const unsigned int *uiSums=(const unsigned int *)hostDst_sums;
int *counts=hostDst_counts;
for(int bin=0;bin<binCount;++bin)
  {
  const int count=counts[bin];
  const unsigned int uiSum=uiSums[bin];
  const double sum=bin+(uiSum*fractFactor)/ONE_IF_ZERO(count);
  sums[bin]=(float)(count*exp((log_dtoaRatio*sum)+log_minDtoa));
  }
return true;
}

//----------------------------------------------------------------------------
