//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuService_ID_Filter1D
{
GLuint program;
int groupSize;
int groupCount;
} GpuService_ID_Filter1D;

static
bool // success
buildProgram(GpuElint *ge,
             GpuService_ID_Filter1D *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " WRITE_ONLY_BUFFER(0, int   backend[]; ) \n"
  "  READ_ONLY_BUFFER(1, int   indices[]; ) \n"
  "  READ_ONLY_BUFFER(2, float values[];  ) \n"
  "                                         \n"
  " UNIFORM_PARAMETER(0, int  )             \n"
  " const int pulseCount=UNI_0.x;           \n"
  "                                         \n"
  " UNIFORM_PARAMETER(1, vec2 )             \n"
  " const float minValue=UNI_1.x,           \n"
  "             maxValue=UNI_1.y;           \n"
  "                                         \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                             \n"
  " main(void) // ID_filter1D()                      \n"
  " {                                                \n"
  " const int highMask=1<<31;                        \n"
  " const bool wrap=minValue>maxValue;               \n"
  " for(int id=globalId;id<pulseCount;id+=GRID_SIZE) \n"
  "   {                                              \n"
  "   int index=indices[id];                         \n"
  "   const float value=values[index];               \n"
  "   const bool c1=value>=minValue;                 \n"
  "   const bool c2=value<=maxValue;                 \n"
  "   const bool keep=wrap ? (c1||c2) : (c1&&c2);    \n"
  "   if(!keep) { index|=highMask; }                 \n"
  "   backend[id]=index;                             \n"
  "   }                                              \n"
  " }                                                \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_ID_Filter1D_init(GpuElint *ge)
{
GpuService_ID_Filter1D *s=Memory_alloc(1,GpuService_ID_Filter1D);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->ID_filter1D_service=s;
return true;
}

void
GpuService_ID_Filter1D_destroy(GpuElint *ge)
{
GpuService_ID_Filter1D *s=ge->ID_filter1D_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

int // number of remaining pulses (-1 on failure)
GpuElint_ID_filter1D(GpuElint *ge,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_values,
                     float minValue,
                     float maxValue)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1; });
const GpuService_ID_Filter1D *s=ge->ID_filter1D_service;
int gpu_backend=GpuElint_ID_compactionBackend(ge,pulseCount);
ON_ERROR(!gpu_backend,"GpuElint_ID_compactionBackend() failure",
         { return -1; });
glUseProgram(s->program);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpu_backend);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_indices);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_values);
ON_GL_ERROR({ return -1; });
glUniform1i(0,pulseCount);
ON_GL_ERROR({ return -1; });
glUniform2f(1,minValue,maxValue);
ON_GL_ERROR({ return -1; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return -1; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return -1; });
const int remaining=GpuElint_ID_compact(ge,gpu_backend,pulseCount,
                                        gpuDst_indices);
ON_ERROR(remaining==-1,"GpuElint_ID_compact() failure",{ return -1; });
return remaining;
}

//----------------------------------------------------------------------------
