//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuService_ID_Filter3D
{
GLuint program;
int groupSize;
int groupCount;
} GpuService_ID_Filter3D;

static
bool // success
buildProgram(GpuElint *ge,
             GpuService_ID_Filter3D *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " WRITE_ONLY_BUFFER(0, int   backend[]; ) \n"
  "  READ_ONLY_BUFFER(1, int   indices[]; ) \n"
  "  READ_ONLY_BUFFER(2, float valuesA[]; ) \n"
  "  READ_ONLY_BUFFER(3, float valuesB[]; ) \n"
  "  READ_ONLY_BUFFER(4, float valuesC[]; ) \n"
  "                                         \n"
  " UNIFORM_PARAMETER(0, int  )             \n"
  " const int pulseCount=UNI_0.x;           \n"
  "                                         \n"
  " UNIFORM_PARAMETER(1, vec3 )             \n"
  " const float minValueA=UNI_1.x,          \n"
  "             minValueB=UNI_1.y,          \n"
  "             minValueC=UNI_1.z;          \n"
  "                                         \n"
  " UNIFORM_PARAMETER(2, vec3 )             \n"
  " const float maxValueA=UNI_2.x,          \n"
  "             maxValueB=UNI_2.y,          \n"
  "             maxValueC=UNI_2.z;          \n"
  "                                         \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                                  \n"
  " main(void) // ID_filter3D()                           \n"
  " {                                                     \n"
  " const int highMask=1<<31;                             \n"
  " const bool wrapA=minValueA>maxValueA;                 \n"
  " const bool wrapB=minValueB>maxValueB;                 \n"
  " const bool wrapC=minValueC>maxValueC;                 \n"
  " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)      \n"
  "   {                                                   \n"
  "   int index=indices[id];                              \n"
  "   const float valueA=valuesA[index];                  \n"
  "   const float valueB=valuesB[index];                  \n"
  "   const float valueC=valuesC[index];                  \n"
  "   const bool c1A=valueA>=minValueA;                   \n"
  "   const bool c2A=valueA<=maxValueA;                   \n"
  "   const bool c1B=valueB>=minValueB;                   \n"
  "   const bool c2B=valueB<=maxValueB;                   \n"
  "   const bool c1C=valueC>=minValueC;                   \n"
  "   const bool c2C=valueC<=maxValueC;                   \n"
  "   const bool keep=(wrapA ? (c1A||c2A) : (c1A&&c2A))&& \n"
  "                   (wrapB ? (c1B||c2B) : (c1B&&c2B))&& \n"
  "                   (wrapC ? (c1C||c2C) : (c1C&&c2C));  \n"
  "   if(!keep) { index|=highMask; }                      \n"
  "   backend[id]=index;                                  \n"
  "   }                                                   \n"
  " }                                                     \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_ID_Filter3D_init(GpuElint *ge)
{
GpuService_ID_Filter3D *s=Memory_alloc(1,GpuService_ID_Filter3D);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->ID_filter3D_service=s;
return true;
}

void
GpuService_ID_Filter3D_destroy(GpuElint *ge)
{
GpuService_ID_Filter3D *s=ge->ID_filter3D_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

int // number of remaining pulses (-1 on failure)
GpuElint_ID_filter3D(GpuElint *ge,
                     int gpuSrc_indices,
                     int pulseCount,
                     int gpuDst_indices,
                     int gpuSrc_valuesA,
                     float minValueA,
                     float maxValueA,
                     int gpuSrc_valuesB,
                     float minValueB,
                     float maxValueB,
                     int gpuSrc_valuesC,
                     float minValueC,
                     float maxValueC)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1; });
const GpuService_ID_Filter3D *s=ge->ID_filter3D_service;
int gpu_backend=GpuElint_ID_compactionBackend(ge,pulseCount);
ON_ERROR(!gpu_backend,"GpuElint_ID_compactionBackend() failure",
         { return -1; });
glUseProgram(s->program);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpu_backend);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_indices);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_valuesA);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,3,gpuSrc_valuesB);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,4,gpuSrc_valuesC);
ON_GL_ERROR({ return -1; });
glUniform1i(0,pulseCount);
ON_GL_ERROR({ return -1; });
glUniform3f(1,minValueA,minValueB,minValueC);
ON_GL_ERROR({ return -1; });
glUniform3f(2,maxValueA,maxValueB,maxValueC);
ON_GL_ERROR({ return -1; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return -1; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return -1; });
const int remaining=GpuElint_ID_compact(ge,gpu_backend,pulseCount,
                                        gpuDst_indices);
ON_ERROR(remaining==-1,"GpuElint_ID_compact() failure",{ return -1; });
return remaining;
}

//----------------------------------------------------------------------------
