//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuService_ID_Histogram1D
{
GLuint program;
int groupSize;
int groupCount;
} GpuService_ID_Histogram1D;

static
bool // success
buildProgram(GpuElint *ge,
             GpuService_ID_Histogram1D *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  "   VOLATILE_BUFFER(0, int   counts[];  ) \n"
  "  READ_ONLY_BUFFER(1, int   indices[]; ) \n"
  "  READ_ONLY_BUFFER(2, float values[];  ) \n"
  "                                         \n"
  " UNIFORM_PARAMETER(0, ivec2 )            \n"
  " const int pulseCount=UNI_0.x,           \n"
  "           binCount  =UNI_0.y;           \n"
  "                                         \n"
  " UNIFORM_PARAMETER(1, vec2  )            \n"
  " const float factor   =UNI_1.x,          \n"
  "             increment=UNI_1.y;          \n"
  "                                         \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                              \n"
  " main(void) // ID_histogram1D()                    \n"
  " {                                                 \n"
  " const int lastBin=binCount-1;                     \n"
  " for(int id=globalId;id<pulseCount;id+=GRID_SIZE)  \n"
  "   {                                               \n"
  "   const int index=indices[id];                    \n"
  "   const float value=values[index];                \n"
  "   const float pos=fma(value,factor,increment);    \n"
  "   const int bin=clamp(int(floor(pos)),0,lastBin); \n"
  "   atomicAdd(counts[bin],1);                       \n"
  "   }                                               \n"
  " }                                                 \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuService_ID_Histogram1D_init(GpuElint *ge)
{
GpuService_ID_Histogram1D *s=Memory_alloc(1,GpuService_ID_Histogram1D);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->ID_histogram1D_service=s;
return true;
}

void
GpuService_ID_Histogram1D_destroy(GpuElint *ge)
{
GpuService_ID_Histogram1D *s=ge->ID_histogram1D_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

bool // success
GpuElint_ID_histogram1D(GpuElint *ge,
                        int gpuSrc_indices,
                        int pulseCount,
                        int gpuDst_counts,
                        int gpuSrc_values,
                        float minValue,
                        float maxValue,
                        int binCount)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
const GpuService_ID_Histogram1D *s=ge->ID_histogram1D_service;
// binCount*(v-minValue)/(maxValue-minValue)
const float factor=binCount/(maxValue-minValue);
const float increment=-minValue*factor;
bool success=GpuElint_zeroFill(ge,gpuDst_counts,0,binCount);
ON_ERROR(!success,"GpuElint_zeroFill() failure",{ return false; });
glUseProgram(s->program);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpuDst_counts);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpuSrc_indices);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_values);
ON_GL_ERROR({ return false; });
glUniform2i(0,pulseCount,binCount);
ON_GL_ERROR({ return false; });
glUniform2f(1,factor,increment);
ON_GL_ERROR({ return false; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return false; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return false; });
return true;
}

//----------------------------------------------------------------------------
