//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuServiceCountFill
{
GLuint program;
int groupSize;
int groupCount;
} GpuServiceCountFill;

static
bool // success
buildProgram(GpuElint *ge,
             GpuServiceCountFill *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                           s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " WRITE_ONLY_BUFFER(0, int outputData[]; ) \n"
  "                                          \n"
  " UNIFORM_PARAMETER(0, ivec4 )             \n"
  " const int begin=UNI_0.x,                 \n"
  "           end=UNI_0.y,                   \n"
  "           countStart=UNI_0.z,            \n"
  "           countIncrement=UNI_0.w;        \n"
  "                                          \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                              \n"
  " main(void) // countFill()                         \n"
  " {                                                 \n"
  " const int offset=countStart-countIncrement*begin; \n"
  " for(int id=begin+globalId;id<end;id+=GRID_SIZE)   \n"
  "   {                                               \n"
  "   outputData[id]=countIncrement*id+offset;        \n"
  "   }                                               \n"
  " }                                                 \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuServiceCountFill_init(GpuElint *ge)
{
GpuServiceCountFill *s=Memory_alloc(1,GpuServiceCountFill);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
ge->countFill_service=s;
return true;
}

void
GpuServiceCountFill_destroy(GpuElint *ge)
{
GpuServiceCountFill *s=ge->countFill_service;
if(!s) return;
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

bool // success
GpuElint_countFill(GpuElint *ge,
                   int gpuDst,
                   int begin,
                   int end,
                   int countStart,
                   int countIncrement)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return false; });
const GpuServiceCountFill *s=ge->countFill_service;
glUseProgram(s->program);
ON_GL_ERROR({ return false; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,gpuDst);
ON_GL_ERROR({ return false; });
glUniform4i(0,begin,end,countStart,countIncrement);
ON_GL_ERROR({ return false; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return false; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return false; });
return true;
}

bool // success
GpuElint_zeroFill(GpuElint *ge,
                  int gpuDst,
                  int begin,
                  int end)
{
return GpuElint_countFill(ge,gpuDst,begin,end,0,0);
}

//----------------------------------------------------------------------------
