//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint_internals.h"

typedef struct _GpuServiceModes4D
{
GLuint program;
int groupSize;
int groupCount;
int gpu_resultCount;
} GpuServiceModes4D;

static
bool // success
buildProgram(GpuElint *ge,
             GpuServiceModes4D *s)
{
char *src=(char *)0;
int srcSize=0;
int srcCapacity=0;
s->groupSize=GpuElint_chooseWorkGroupSize(ge,false);
s->groupCount=GpuElint_chooseWorkGroupCount(ge);
GpuElint_appendCodePreamble(ge,&src,&srcSize,&srcCapacity,
                            s->groupSize,s->groupCount);
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " struct HistoMode4D                            \n"
  " {                                             \n"
  " int maxCount,weight;                          \n"
  " int firstBinA,lastBinA;                       \n"
  " int firstBinB,lastBinB;                       \n"
  " int firstBinC,lastBinC;                       \n"
  " int firstBinD,lastBinD;                       \n"
  " };                                            \n"
  "                                               \n"
  "   VOLATILE_BUFFER(0, int resultCount;       ) \n"
  " WRITE_ONLY_BUFFER(1, HistoMode4D results[]; ) \n"
  "  READ_ONLY_BUFFER(2, int counts[];          ) \n"
  "                                               \n"
  " UNIFORM_PARAMETER(0, float )                  \n"
  " const float cutoffRatio=UNI_0.x;              \n"
  "                                               \n"
  " UNIFORM_PARAMETER(1, int   )                  \n"
  " const int modeCapacity=UNI_1.x;               \n"
  "                                               \n"
  " UNIFORM_PARAMETER(2, ivec4 )                  \n"
  " const int binCountA=UNI_2.x,                  \n"
  "           binCountB=UNI_2.y,                  \n"
  "           binCountC=UNI_2.z,                  \n"
  "           binCountD=UNI_2.w;                  \n"
  "                                               \n"
  " UNIFORM_PARAMETER(3, ivec4 )                  \n"
  " const bool cyclicA=UNI_3.x!=0,                \n"
  "            cyclicB=UNI_3.y!=0,                \n"
  "            cyclicC=UNI_3.z!=0,                \n"
  "            cyclicD=UNI_3.w!=0;                \n"
  "                                               \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  " void                                                     \n"
  " main(void) // modes4D()                                  \n"
  " {                                                        \n"
  " const int binCount=binCountA*binCountB;                  \n"
  " for(int bin=globalId;bin<binCount;bin+=GRID_SIZE)        \n"
  "   {                                                      \n"
  "   const int binA=bin%%binCountA;                         \n"
  "   const int binB=(bin/binCountA)%%binCountB;             \n"
  "   const int binC=(bin/(binCountA*binCountB))%%binCountC; \n"
  "   const int binD=bin/(binCountA*binCountB*binCountC);    \n"
  "   int count=counts[bin];                                 \n"
  "   if(count==0) { continue; }                             \n"
  "   HistoMode4D m={ count, count,                          \n"
  "                   binA, binA, binB, binB,                \n"
  "                   binC, binC, binD, binD };              \n"
  "   const int cutoff=int(floor(cutoffRatio*m.maxCount));   \n"
  "   bool giveUp=false;                                     \n"
  "   int stop=cyclicA ? binA : binCountA-1;                 \n"
  "   int prevCount=m.maxCount;                              \n"
  "   bool lowEnough=false;                                  \n"
  "   for(int i=binA;(i=(i+binCountA-1)%%binCountA)!=stop;)  \n"
  "     {                                                    \n"
  "     count=counts[((binD*binCountC+binC)*binCountB+binB)* \n"
  "                  binCountA+i];                           \n"
  "     if(count>m.maxCount) { giveUp=true; break; }         \n"
  "     if(count<=cutoff) { lowEnough=true; }                \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))      \n"
  "       { break; }                                         \n"
  "     m.weight+=count;                                     \n"
  "     m.firstBinA=i;                                       \n"
  "     prevCount=count;                                     \n"
  "     }                                                    \n"
  "   if(giveUp) { continue; }                               \n"
  "   stop=cyclicA ? binA : 0;                               \n"
  "   prevCount=m.maxCount;                                  \n"
  "   lowEnough=false;                                       \n"
  "   for(int i=binA;(i=(i+1)%%binCountA)!=stop;)            \n"
  "     {                                                    \n"
  "     count=counts[((binD*binCountC+binC)*binCountB+binB)* \n"
  "                  binCountA+i];                           \n"
  "     if(count>m.maxCount) { giveUp=true; break; }         \n"
  "     if(count<=cutoff) { lowEnough=true; }                \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))      \n"
  "       { break; }                                         \n"
  "     m.weight+=count;                                     \n"
  "     m.lastBinA=i;                                        \n"
  "     prevCount=count;                                     \n"
  "     }                                                    \n"
  "   if(giveUp) { continue; }                               \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  "   stop=cyclicB ? binB : binCountB-1;                    \n"
  "   prevCount=m.maxCount;                                 \n"
  "   lowEnough=false;                                      \n"
  "   for(int i=binB;(i=(i+binCountB-1)%%binCountB)!=stop;) \n"
  "     {                                                   \n"
  "     count=counts[((binD*binCountC+binC)*binCountB+i)*   \n"
  "                  binCountA+binA];                       \n"
  "     if(count>m.maxCount) { giveUp=true; break; }        \n"
  "     if(count<=cutoff) { lowEnough=true; }               \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))     \n"
  "       { break; }                                        \n"
  "     m.weight+=count;                                    \n"
  "     m.firstBinB=i;                                      \n"
  "     prevCount=count;                                    \n"
  "     }                                                   \n"
  "   if(giveUp) { continue; }                              \n"
  "   stop=cyclicB ? binB : 0;                              \n"
  "   prevCount=m.maxCount;                                 \n"
  "   lowEnough=false;                                      \n"
  "   for(int i=binB;(i=(i+1)%%binCountB)!=stop;)           \n"
  "     {                                                   \n"
  "     count=counts[((binD*binCountC+binC)*binCountB+i)*   \n"
  "                  binCountA+binA];                       \n"
  "     if(count>m.maxCount) { giveUp=true; break; }        \n"
  "     if(count<=cutoff) { lowEnough=true; }               \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))     \n"
  "       { break; }                                        \n"
  "     m.weight+=count;                                    \n"
  "     m.lastBinB=i;                                       \n"
  "     prevCount=count;                                    \n"
  "     }                                                   \n"
  "   if(giveUp) { continue; }                              \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  "   stop=cyclicC ? binC : binCountC-1;                    \n"
  "   prevCount=m.maxCount;                                 \n"
  "   lowEnough=false;                                      \n"
  "   for(int i=binC;(i=(i+binCountC-1)%%binCountC)!=stop;) \n"
  "     {                                                   \n"
  "     count=counts[((binD*binCountC+i)*binCountB+binB)*   \n"
  "                  binCountA+binA];                       \n"
  "     if(count>m.maxCount) { giveUp=true; break; }        \n"
  "     if(count<=cutoff) { lowEnough=true; }               \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))     \n"
  "       { break; }                                        \n"
  "     m.weight+=count;                                    \n"
  "     m.firstBinC=i;                                      \n"
  "     prevCount=count;                                    \n"
  "     }                                                   \n"
  "   if(giveUp) { continue; }                              \n"
  "   stop=cyclicC ? binC : 0;                              \n"
  "   prevCount=m.maxCount;                                 \n"
  "   lowEnough=false;                                      \n"
  "   for(int i=binC;(i=(i+1)%%binCountC)!=stop;)           \n"
  "     {                                                   \n"
  "     count=counts[((binD*binCountC+i)*binCountB+binB)*   \n"
  "                  binCountA+binA];                       \n"
  "     if(count>m.maxCount) { giveUp=true; break; }        \n"
  "     if(count<=cutoff) { lowEnough=true; }               \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))     \n"
  "       { break; }                                        \n"
  "     m.weight+=count;                                    \n"
  "     m.lastBinC=i;                                       \n"
  "     prevCount=count;                                    \n"
  "     }                                                   \n"
  "   if(giveUp) { continue; }                              \n");
GpuElint_appendText(ge,&src,&srcSize,&srcCapacity,
  "   stop=cyclicD ? binD : binCountD-1;                    \n"
  "   prevCount=m.maxCount;                                 \n"
  "   lowEnough=false;                                      \n"
  "   for(int i=binD;(i=(i+binCountD-1)%%binCountD)!=stop;) \n"
  "     {                                                   \n"
  "     count=counts[((i*binCountC+binC)*binCountB+binB)*   \n"
  "                  binCountA+binA];                       \n"
  "     if(count>m.maxCount) { giveUp=true; break; }        \n"
  "     if(count<=cutoff) { lowEnough=true; }               \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))     \n"
  "       { break; }                                        \n"
  "     m.weight+=count;                                    \n"
  "     m.firstBinD=i;                                      \n"
  "     prevCount=count;                                    \n"
  "     }                                                   \n"
  "   if(giveUp) { continue; }                              \n"
  "   stop=cyclicD ? binD : 0;                              \n"
  "   prevCount=m.maxCount;                                 \n"
  "   lowEnough=false;                                      \n"
  "   for(int i=binD;(i=(i+1)%%binCountD)!=stop;)           \n"
  "     {                                                   \n"
  "     count=counts[((i*binCountC+binC)*binCountB+binB)*   \n"
  "                  binCountA+binA];                       \n"
  "     if(count>m.maxCount) { giveUp=true; break; }        \n"
  "     if(count<=cutoff) { lowEnough=true; }               \n"
  "     if((count==0)||((count>=prevCount)&&lowEnough))     \n"
  "       { break; }                                        \n"
  "     m.weight+=count;                                    \n"
  "     m.lastBinD=i;                                       \n"
  "     prevCount=count;                                    \n"
  "     }                                                   \n"
  "   if(giveUp) { continue; }                              \n"
  "   const int pos=atomicAdd(resultCount,1);               \n"
  "   if(pos<modeCapacity) { results[pos]=m; }              \n"
  "   }                                                     \n"
  " }                                                       \n");
s->program=GpuElint_buildProgram(ge,src);
Memory_free(src);
return s->program!=0;
}

bool // success
GpuServiceModes4D_init(GpuElint *ge)
{
GpuServiceModes4D *s=Memory_alloc(1,GpuServiceModes4D);
if(!buildProgram(ge,s))
  {
  Memory_free(s);
  return false;
  }
s->gpu_resultCount=GpuElint_gpuAlloc(ge,sizeof(int32_t));
ON_ERROR(!s->gpu_resultCount,"GpuElint_gpuAlloc() failure",
         {
         glDeleteProgram(s->program);
         Memory_free(s);
         return false;
         });
ge->modes4D_service=s;
return true;
}

void
GpuServiceModes4D_destroy(GpuElint *ge)
{
GpuServiceModes4D *s=ge->modes4D_service;
if(!s) return;
bool success=GpuElint_gpuFree(ge,s->gpu_resultCount);
ON_ERROR(!success,"GpuElint_gpuFree() failure",{});
glDeleteProgram(s->program);
ON_GL_ERROR({});
Memory_free(s);
}

int // number of modes (-1 on failure)
GpuElint_modes4D(GpuElint *ge,
                 HistoMode4D *hostDst_modes,
                 int modeCapacity,
                 int gpuSrc_counts,
                 float cutoffRatio,
                 int binCountA,
                 bool cyclicA,
                 int binCountB,
                 bool cyclicB,
                 int binCountC,
                 bool cyclicC,
                 int binCountD,
                 bool cyclicD)
{
GPUELINT_API_INVOCATION_CHECK(ge,{ return -1; });
const GpuServiceModes4D *s=ge->modes4D_service;
int gpu_results=GpuElint_gpuBackend(ge,modeCapacity*sizeof(HistoMode4D));
ON_ERROR(!gpu_results,"GpuElint_gpuBackend() failure",{ return -1; });
int32_t modeCount=0;
bool success=GpuElint_hostToGpu(ge,s->gpu_resultCount,&modeCount,
                                0,0,sizeof(int32_t));
ON_ERROR(!success,"GpuElint_hostToGpu() failure",{ return -1; });
glUseProgram(s->program);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,0,s->gpu_resultCount);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,1,gpu_results);
ON_GL_ERROR({ return -1; });
glBindBufferBase(GL_SHADER_STORAGE_BUFFER,2,gpuSrc_counts);
ON_GL_ERROR({ return -1; });
glUniform1f(0,cutoffRatio);
ON_GL_ERROR({ return -1; });
glUniform1i(1,modeCapacity);
ON_GL_ERROR({ return -1; });
glUniform4i(2,binCountA,binCountB,binCountC,binCountD);
ON_GL_ERROR({ return -1; });
glUniform4i(3,cyclicA,cyclicB,cyclicC,cyclicD);
ON_GL_ERROR({ return -1; });
glDispatchCompute(s->groupCount,1,1);
ON_GL_ERROR({ return -1; });
glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);
ON_GL_ERROR({ return -1; });
success=GpuElint_gpuToHost(ge,&modeCount,s->gpu_resultCount,
                           0,0,sizeof(int32_t));
ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return -1; });
if(modeCount)
  {
  const int copyCount=modeCount>modeCapacity ? modeCapacity : modeCount;
  success=GpuElint_gpuToHost(ge,hostDst_modes,gpu_results,
                             0,0,copyCount*sizeof(HistoMode4D));
  ON_ERROR(!success,"GpuElint_gpuToHost() failure",{ return -1; });
  // FIXME: if the maximal count appears many times in a mode, this mode
  //        will then be reported many times.
  //        Would it be worth filtering to remove duplicates?
  }
return modeCount;
}

//----------------------------------------------------------------------------
