//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

import java.awt.Frame;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;

import javax.media.opengl.awt.GLCanvas;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GL4;
import javax.media.opengl.GLProfile;
import javax.media.opengl.GLCapabilities;

import com.jogamp.opengl.util.Animator;

import java.io.FileInputStream;
import java.nio.channels.FileChannel;

import java.nio.ByteBuffer;
import java.nio.IntBuffer;
import java.nio.FloatBuffer;
import java.nio.DoubleBuffer;

//----------------------------------------------------------------------------
// Application-specific code
//----------------------------------------------------------------------------

class Application
{
static final int HISTO_A=0;
static final int HISTO_B=3;

static final String[] propertyName ={ "DI", "FN", "NIV", "AZIMUT" };
static final int[] propertyBinCount={ 128,  32,   32,    128      };
static final String NAME_A=propertyName[HISTO_A];
static final String NAME_B=propertyName[HISTO_B];
static final int BIN_COUNT_A=propertyBinCount[HISTO_A];
static final int BIN_COUNT_B=propertyBinCount[HISTO_B];

static final int MAX_HISTO_MODE_COUNT=128;
static final float HISTO_CUTOFF_RATIO=0.1f;
static final float HIGHEST_PRF=100000.0f;
static final float LOWEST_PRF=100.0f;
static final float DTOA_RATIO=1.025f;
static final float DTOA_DETECTION_THRESHOLD=0.01f;
static final int MIN_MATCH_COUNT=4;

static final int sizeof_double=Double.SIZE/8;
static final int sizeof_float=Float.SIZE/8;
static final int sizeof_int=Integer.SIZE/8;

//-- properties set by execution environment --
GLCanvas canvas=null;
int width=0,height=0;
String fileName=null;

//-- simple pass-through vertex-shader source code --
String vertexSrc=
" #version 430                                                            \n"+
" in vec2 v_Vertex;                                                       \n"+
" in int v_Mark;                                                          \n"+
" flat out int f_Mark;                                                    \n"+
" uniform vec2 u_Delta;                                                   \n"+
" void                                                                    \n"+
" main(void)                                                              \n"+
" {                                                                       \n"+
" f_Mark=v_Mark;                                                          \n"+
" gl_Position=vec4(v_Vertex+u_Delta,0.0,1.0);                             \n"+
" }                                                                       \n";

//-- simple unicolour fragment-shader source code --
String fragmentSrc=
" #version 430                                                            \n"+
" layout(location=0) out vec4 defaultFragColour;                          \n"+
" #define gl_FragColor defaultFragColour                                  \n"+
" flat in int f_Mark;                                                     \n"+
" void                                                                    \n"+
" main(void)                                                              \n"+
" {                                                                       \n"+
" gl_FragColor=f_Mark<0 ? vec4(0.2,0.8,0.2,1.0)                           \n"+
"                       : vec4(0.8,0.2,0.2,1.0);                          \n"+
" }                                                                       \n";

//-- shader-program and parameters --
int program;
int v_Vertex;
int v_Mark;
int u_Delta;

//-- informations required to draw a simple shape --
int vertexCount;
int vertex_vbo;
int mark_vbo;
int vao;
volatile int stepCount;
volatile int redrawCount;

//-- GPU execution environment --
JavaGpuElint ge;

boolean // success
init()
  {
  GL4 gl=canvas.getGL().getGL4();
  try
    {
    ge=new JavaGpuElint(null,true);
    }
  catch(Throwable t)
    {
    System.err.printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    System.err.printf("~~~~ GpuElint initialisation failure ~~~~\n");
    t.printStackTrace();
    System.err.printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    return false;
    }
  try
    {
    JavaGpuElint.Properties properties=ge.properties();
    System.out.printf("DTOA_SKIP_COUNT=%d\n",
                       properties.DTOA_SKIP_COUNT);
    System.out.printf("MAX_MATCH_COUNT=%d\n",
                       properties.MAX_MATCH_COUNT);
    System.out.printf("glVersionMajor=%d\n",
                       properties.glVersionMajor);
    System.out.printf("glVersionMinor=%d\n",
                       properties.glVersionMinor);
    System.out.printf("glslVersion=%d\n",
                       properties.glslVersion);
    System.out.printf("maxWorkGroupCountX=%d\n",
                       properties.maxWorkGroupCountX);
    System.out.printf("maxWorkGroupCountY=%d\n",
                       properties.maxWorkGroupCountY);
    System.out.printf("maxWorkGroupCountZ=%d\n",
                       properties.maxWorkGroupCountZ);
    System.out.printf("maxWorkGroupSizeX=%d\n",
                       properties.maxWorkGroupSizeX);
    System.out.printf("maxWorkGroupSizeY=%d\n",
                       properties.maxWorkGroupSizeY);
    System.out.printf("maxWorkGroupSizeZ=%d\n",
                       properties.maxWorkGroupSizeZ);
    System.out.printf("maxWorkGroupInvocations=%d\n",
                       properties.maxWorkGroupInvocations);
    System.out.printf("maxUniformBlocks=%d\n",
                       properties.maxUniformBlocks);
    System.out.printf("maxTextureImageUnits=%d\n",
                       properties.maxTextureImageUnits);
    System.out.printf("maxAtomicCounterBuffers=%d\n",
                       properties.maxAtomicCounterBuffers);
    System.out.printf("maxAtomicCounters=%d\n",
                       properties.maxAtomicCounters);
    System.out.printf("maxSharedMemorySize=%d\n",
                       properties.maxSharedMemorySize);
    System.out.printf("maxUniformComponents=%d\n",
                       properties.maxUniformComponents);
    System.out.printf("maxImageUniforms=%d\n",
                       properties.maxImageUniforms);
    System.out.printf("maxCombinedUniformComponents=%d\n",
                       properties.maxCombinedUniformComponents);
    System.out.printf("maxShaderStorageBlocks=%d\n",
                       properties.maxShaderStorageBlocks);
    System.out.printf("warpSize=%d\n",
                       properties.warpSize);
    System.out.printf("warpsPerStreamingMultiprocessor=%d\n",
                       properties.warpsPerStreamingMultiprocessor);
    System.out.printf("streamingMultiprocessorCount=%d\n",
                       properties.streamingMultiprocessorCount);
    System.out.printf("hasBallot=%b\n",
                       properties.hasBallot);
    System.out.printf("hasShuffle=%b\n",
                       properties.hasShuffle);

    //-- detect platform-specific features --
    System.out.printf("OpenGL version: %s\n",
                      gl.glGetString(GL4.GL_VERSION));
    System.out.printf("          GLSL: %s\n",
                      gl.glGetString(GL4.GL_SHADING_LANGUAGE_VERSION));
    System.out.printf("        vendor: %s\n",
                      gl.glGetString(GL4.GL_VENDOR));
    System.out.printf("      renderer: %s\n",
                      gl.glGetString(GL4.GL_RENDERER));

    //-- load data from file --
    final int pulseCount=
      loadData(ge,fileName,0,sizeof_int).asIntBuffer().get(0);
    final DoubleBuffer host_toa64=
      loadData(ge,fileName,
               sizeof_int,
               pulseCount*sizeof_double).asDoubleBuffer();
    final FloatBuffer host_valuesA=
      loadData(ge,fileName,
               sizeof_int+
               pulseCount*sizeof_double+
               pulseCount*sizeof_float*HISTO_A,
               pulseCount*sizeof_float).asFloatBuffer();
    final FloatBuffer host_valuesB=
      loadData(ge,fileName,
               sizeof_int+
               pulseCount*sizeof_double+
               pulseCount*sizeof_float*HISTO_B,
               pulseCount*sizeof_float).asFloatBuffer();

    //-- change initial value for some OpenGL settings --
    gl.glClearColor(0.3f,0.3f,0.3f,1.0f);

    //-- create shader-program and access its parameters --
    int vs=gl.glCreateShader(GL4.GL_VERTEX_SHADER);
    String[] vsStrings={ vertexSrc };
    gl.glShaderSource(vs,1,vsStrings,null,0);
    gl.glCompileShader(vs);
    int fs=gl.glCreateShader(GL4.GL_FRAGMENT_SHADER);
    String[] fsStrings={ fragmentSrc };
    gl.glShaderSource(fs,1,fsStrings,null,0);
    gl.glCompileShader(fs);
    program=gl.glCreateProgram();
    gl.glAttachShader(program,vs);
    gl.glAttachShader(program,fs);
    gl.glLinkProgram(program);
    gl.glDeleteShader(vs);
    gl.glDeleteShader(fs);
    v_Vertex=gl.glGetAttribLocation(program,"v_Vertex");
    v_Mark=gl.glGetAttribLocation(program,"v_Mark");
    u_Delta=gl.glGetUniformLocation(program,"u_Delta");

    //-- compute vertices --
    vertexCount=pulseCount;
    FloatBuffer tmpV=ge.hostAlloc(vertexCount*2*sizeof_float)
                       .asFloatBuffer();
    int side=(int)Math.sqrt(vertexCount+1.0);
    double factor=1.8/side;
    double increment=-0.9;
    for(int i=0;i<vertexCount;++i)
      {
      tmpV.put(2*i+0,(float)(factor*(i%side)+increment));
      tmpV.put(2*i+1,(float)(factor*(i/side)+increment));
      }

    //-- create and fill buffers with vertices --
    int[] out_vbo={ 0 };
    gl.glGenBuffers(1,out_vbo,0);
    vertex_vbo=out_vbo[0];
    gl.glBindBuffer(GL4.GL_ARRAY_BUFFER,vertex_vbo);
    gl.glBufferData(GL4.GL_ARRAY_BUFFER,
                    vertexCount*2*sizeof_float,tmpV,
                    GL4.GL_STATIC_DRAW);
    mark_vbo=ge.gpuAlloc(vertexCount*sizeof_int);
    ge.countFill(mark_vbo,0,vertexCount,-1,0);

    //-- create and set vertex-array up --
    int[] out_vao={ 0 };
    gl.glGenVertexArrays(1,out_vao,0);
    vao=out_vao[0];
    gl.glBindVertexArray(vao);
    //-- associate vertex-array attribute and buffer --
    gl.glEnableVertexAttribArray(v_Vertex);
    gl.glBindBuffer(GL4.GL_ARRAY_BUFFER,vertex_vbo);
    gl.glVertexAttribPointer(v_Vertex,2,GL4.GL_FLOAT,false,0,0);
    gl.glEnableVertexAttribArray(v_Mark);
    gl.glBindBuffer(GL4.GL_ARRAY_BUFFER,mark_vbo);
    gl.glVertexAttribPointer(v_Mark,1,GL4.GL_INT,false,0,0);
    gl.glBindVertexArray(0); // avoid accidental change

    //-- start an asynchronous computation thread if requested --
    gl.glFinish(); //  ensure graphics initialisation is done
    stepCount=0;
    redrawCount=0;
    System.out.printf("----> launching an asynchronous computation thread\n");
    Thread th=new Thread()
      {
      @Override public void run()
        {
        try
          {
          computeThread(pulseCount,host_toa64,host_valuesA,host_valuesB);
          }
        catch(Throwable t)
          {
          // What the hell should I do with this f.....g exception?
          // run() can't even throw it further!
          t.printStackTrace();
          }
        }
      };
    th.setDaemon(true);
    th.start();
    }
  catch(Throwable t)
    {
    t.printStackTrace();
    return false;
    }
  return true;
  }

ByteBuffer
loadData(JavaGpuElint ge,
         String fileName,
         long offset,
         long size) throws Throwable
  {
  ByteBuffer buffer=ge.hostAlloc(size);
  FileInputStream f=new FileInputStream(fileName);
  FileChannel c=f.getChannel();
  c.position(offset);
  c.read(buffer);
  buffer.rewind();
  return buffer;
  }

void
computeThread(int pulseCount,
              DoubleBuffer host_toa64,
              FloatBuffer host_valuesA,
              FloatBuffer host_valuesB) throws Throwable
  {
  JavaGpuElint ge=null;
  try
    {
    ge=new JavaGpuElint(this.ge,true);
    }
  catch(Throwable t)
    {
    System.err.printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    System.err.printf("~~~~ shared GpuElint initialisation failure ~~~~\n");
    t.printStackTrace();
    System.err.printf("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
    }
  for(;;)
    {
    for(int i=0;i<100;++i)
      { Thread.sleep(10); step(); }
    if(ge!=null)
      { compute(ge,pulseCount,host_toa64,host_valuesA,host_valuesB); }
    }
  }

void
step()
  {
  synchronized(this) { ++stepCount; this.notifyAll(); }
  canvas.repaint();
  }

int[]
toaAnalysis(JavaGpuElint ge,
            int indices,
            int pulseCount,
            int toa64,
            int marks,
            int seqMark,
            float minDtoa,
            int dtoaBinCount,
            IntBuffer dtoaCounts,
            FloatBuffer dtoaSums,
            float minDuration)
  {
  int prevSeqMark=seqMark;
  float priRatioThreshold=DTOA_RATIO*DTOA_RATIO;
  float exclusivePriLowerLimit=0.0f;
  for(;;)
    {
    if(exclusivePriLowerLimit==0.0f)
      {
      ge.dtoaHistogram(indices,pulseCount,toa64,
                       dtoaCounts,dtoaSums,dtoaBinCount,
                       minDtoa,DTOA_RATIO);
      step();
      }
    float estimatedPri=ge.estimateMajorPri(dtoaCounts,dtoaSums,dtoaBinCount,
                                           minDtoa,DTOA_RATIO,
                                           exclusivePriLowerLimit,minDuration);
    step();
    if(estimatedPri==0.0f) { break; }
    System.out.printf("estimated PRI: %g (%.3f Hz)\n",
                      estimatedPri,1.0f/estimatedPri);
    float pri=ge.markPulses(indices,pulseCount,toa64,
                            marks,seqMark,estimatedPri,
                            priRatioThreshold,MIN_MATCH_COUNT);
    step();
    if(pri>0.0f)
      {
      int remaining=ge.filterUnmarked(indices,pulseCount,indices,marks);
      step();
      int extractedCount=pulseCount-remaining;
      pulseCount=remaining; // don't consider marked pulses anymore
      System.out.printf("[%3d] ---> found %d pulses for %.3f Hz\n",
                        seqMark,extractedCount,1.0f/pri);
      ++seqMark;
      exclusivePriLowerLimit=0.0f;
      }
    else
      { exclusivePriLowerLimit=estimatedPri; }
    }
  return new int[]{ seqMark-prevSeqMark , seqMark };
  }

void
compute(JavaGpuElint ge,
        int pulseCount,
        DoubleBuffer host_toa64,
        FloatBuffer host_valuesA,
        FloatBuffer host_valuesB) throws Throwable
  {
  System.out.printf("#######################################"+
                    "#######################################\n");
  System.out.printf("## gpu analysis of %d pulses\n",pulseCount);

  //-- send data to GPU --
  int toa64=ge.gpuAlloc(pulseCount*sizeof_double);
  ge.hostToGpu(toa64,host_toa64,0,0,pulseCount*sizeof_double);
  int valuesA=ge.gpuAlloc(pulseCount*sizeof_float);
  ge.hostToGpu(valuesA,host_valuesA,0,0,pulseCount*sizeof_float);
  int valuesB=ge.gpuAlloc(pulseCount*sizeof_float);
  ge.hostToGpu(valuesB,host_valuesB,0,0,pulseCount*sizeof_float);
  step();

  //-- prepare ancillary data --
  float minDuration=((float)(host_toa64.get(pulseCount-1)-
                             host_toa64.get(0)))*DTOA_DETECTION_THRESHOLD;
  JavaGpuElint.HistoMode2D[] modes=
    new JavaGpuElint.HistoMode2D[MAX_HISTO_MODE_COUNT];
  int histo=ge.gpuAlloc(BIN_COUNT_A*BIN_COUNT_B*sizeof_int);
  float minDtoa=1.0f/HIGHEST_PRF;
  float maxDtoa=1.0f/LOWEST_PRF;
  int dtoaBinCount=(int)Math.ceil((Math.log(maxDtoa)-Math.log(minDtoa))/
                                  Math.log(DTOA_RATIO));
  IntBuffer dtoaCounts=ge.hostAlloc(dtoaBinCount*sizeof_int)
                         .asIntBuffer();
  FloatBuffer dtoaSums=ge.hostAlloc(dtoaBinCount*sizeof_float)
                         .asFloatBuffer();
  int indices=ge.gpuAlloc(pulseCount*sizeof_int);
  int toaIndices=ge.gpuAlloc(pulseCount*sizeof_int);
  ge.countFill(indices,0,pulseCount,0,1);
  int marks=mark_vbo; // shared with rendering
  ge.countFill(marks,0,pulseCount,-1,0);
  ge.flush(true);
  step();
  int seqMark=0;
  for(;;)
    {
    System.out.printf("---------------------------------------"+
                      "---------------------------------------\n");
    int prevSeqMark=seqMark;
    float minValueA,maxValueA;
    boolean cyclicA=false;
    if(NAME_A.equals("AZIMUT"))
      { minValueA=0.0f; maxValueA=360.0f; cyclicA=true; }
    else
      {
      JavaGpuElint.MinMaxResult r=ge.minMax(indices,pulseCount,valuesA);
      step();
      minValueA=r.minValue;
      maxValueA=r.maxValue;
      }
    float minValueB,maxValueB;
    boolean cyclicB=false;
    if(NAME_B.equals("AZIMUT"))
      { minValueB=0.0f; maxValueB=360.0f; cyclicB=true; }
    else
      {
      JavaGpuElint.MinMaxResult r=ge.minMax(indices,pulseCount,valuesB);
      step();
      minValueB=r.minValue;
      maxValueB=r.maxValue;
      }
    ge.histogram2D(indices,pulseCount,histo,
                   valuesA,minValueA,maxValueA,BIN_COUNT_A,
                   valuesB,minValueB,maxValueB,BIN_COUNT_B);
    step();
    int modeCount=ge.modes2D(modes,histo,HISTO_CUTOFF_RATIO,
                             BIN_COUNT_A,cyclicA,
                             BIN_COUNT_B,cyclicB);
    step();
    System.out.printf("%d modes found in %s/%s histogram\n",
                      modeCount,NAME_A,NAME_B);
    if(modeCount>modes.length)
      {
      System.out.printf("MAX_HISTO_MODE_COUNT exceeded (%d)\n",modeCount);
      modeCount=modes.length;
      }
    step();
    for(;;)
      {
      boolean lastAttempt=(modeCount<=1);
      int toaPulseCount;
      if(lastAttempt) // consider every remaining pulse
        {
        ge.gpuToGpu(toaIndices,indices,0,0,pulseCount*sizeof_int);
        step();
        toaPulseCount=pulseCount;
        System.out.printf("analysing %d remaining pulses\n",toaPulseCount);
        }
      else // consider only the best mode
        {
        int bestMode=0;
        for(int i=1;i<modeCount;++i)
          { if(modes[i].weight>modes[bestMode].weight) { bestMode=i; } }
        float histoStepA=(maxValueA-minValueA)/BIN_COUNT_A;
        float minFilterA=minValueA+(modes[bestMode].firstBinA+0.0f)*histoStepA;
        float maxFilterA=minValueA+(modes[bestMode].lastBinA+1.0f)*histoStepA;
        float histoStepB=(maxValueB-minValueB)/BIN_COUNT_B;
        float minFilterB=minValueB+(modes[bestMode].firstBinB+0.0f)*histoStepB;
        float maxFilterB=minValueB+(modes[bestMode].lastBinB+1.0f)*histoStepB;
        toaPulseCount=ge.filter2D(indices,pulseCount,toaIndices,
                                  valuesA,minFilterA,maxFilterA,
                                  valuesB,minFilterB,maxFilterB);
        step();
        System.out.printf("analysing %d pulses out of %d"+
                      " (%s/%s mode from bin %d/%d to %d/%d)\n",
                      toaPulseCount,pulseCount,NAME_A,NAME_B,
                      modes[bestMode].firstBinA,modes[bestMode].firstBinB,
                      modes[bestMode].lastBinA,modes[bestMode].lastBinB);
        modes[bestMode]=modes[--modeCount]; // forget this mode
        }
      int[] r=toaAnalysis(ge,toaIndices,toaPulseCount,toa64,
                          marks,seqMark,minDtoa,dtoaBinCount,
                          dtoaCounts,dtoaSums,minDuration);
      int seqCount=r[0];
      seqMark=r[1];
      if(seqCount!=0)
        {
        // don't consider marked pulses anymore
        pulseCount=ge.filterUnmarked(indices,pulseCount,indices,marks);
        break; // histogram will be rebuilt
        }
      if(lastAttempt) { break; }
      }
    if(prevSeqMark==seqMark)
      { break; } // nothing was found with this histogram
    }
  //-- cleanup --
  // marks is mark_vbo (shared with rendering)
  ge.gpuFree(toaIndices);
  ge.gpuFree(indices);
  ge.gpuFree(histo);
  ge.gpuFree(valuesB);
  ge.gpuFree(valuesA);
  ge.gpuFree(toa64);
  }

void
redraw() throws Throwable
  {
  GL4 gl=canvas.getGL().getGL4();

  //-- wait for next computation step change --
  synchronized(this)
    {
    while(stepCount==redrawCount) { this.wait(); }
    redrawCount=stepCount;
    }

  //-- animate drawing --
  double a=Math.PI*0.25*0.001*System.currentTimeMillis();
  float dx=0.1f*(float)Math.cos(a);
  float dy=0.1f*(float)Math.sin(a);

  //-- reset framebuffer --
  gl.glClear(GL4.GL_COLOR_BUFFER_BIT);
  gl.glViewport(0,0,width,height);

  //-- draw vertex-array using shader-program --
  gl.glUseProgram(program);
  gl.glUniform2f(u_Delta,dx,dy);
  gl.glBindVertexArray(vao);
  gl.glDrawArrays(GL4.GL_POINTS,0,vertexCount);
  gl.glFinish();
  }

void
keyboard(KeyEvent e)
  {
  }
}

//----------------------------------------------------------------------------
// Execution environment
//----------------------------------------------------------------------------

public class TestGpuElint
{
Application application;
GLCanvas canvas;
Frame frame;
Animator animator;

TestGpuElint(String title,
             String fileName,
             int width,
             int height)
  {
  application=new Application();
  application.fileName=fileName;
  GLProfile profile=GLProfile.getDefault();
  GLCapabilities capabilities=new GLCapabilities(profile);
  // System.out.printf("capabilities: %s\n",capabilities.toString());
  canvas=new GLCanvas(capabilities);
  animator=new Animator(canvas);
  frame=new Frame(title);
  frame.add(canvas);
  frame.setSize(width,height);
  canvas.addGLEventListener(new GLEventListener()
    {
    @Override public void init(GLAutoDrawable d)
      {
      // postpone initialisation until initial size is known
      if((application.canvas==null)&&
         (application.width>0)&&
         (application.height>0))
        {
        application.canvas=canvas;
        if(!application.init()) { System.exit(1); }
        // FIXME: this does not even exit the process! (enjoy...)
        }
      }
    @Override public void reshape(GLAutoDrawable d,int x,int y,int w,int h)
      {
      application.width=w;
      application.height=h;
      // initialisation may have been postponed
      if(application.canvas==null)
        {
        application.canvas=canvas;
        if(!application.init()) { System.exit(1); }
        // FIXME: this does not even exit the process! (enjoy...)
        }
      }
    @Override public void display(GLAutoDrawable d)
      {
      if(application.canvas!=null)
        {
        try
          {
          application.redraw();
          }
        catch(Throwable t)
          {
          // What the hell should I do with this f.....g exception?
          // display() can't even throw it further!
          t.printStackTrace();
          }
        }
      }
    @Override public void dispose(GLAutoDrawable d)
      { /* nothing to be done */ }
    });
  frame.addWindowListener(new WindowAdapter()
    {
    public void windowClosing(WindowEvent e)
      { exit(0); }
    });
  canvas.addKeyListener(new KeyAdapter()
    {
    public void keyPressed(KeyEvent e)
      {
      if(e.getKeyCode()==KeyEvent.VK_ESCAPE) exit(0);
      else application.keyboard(e);
      }
    });
  frame.setVisible(true);
  animator.start();
  canvas.requestFocus();
  }

void
exit(int status)
  {
  animator.stop();
  frame.dispose();
  System.exit(status);
  }

public static
void
main(String[] args)
  {
  if(args.length!=1)
    {
    System.err.printf("usage: TestGpuElint pulse_file.raw\n");
    System.exit(1);
    }
  String fileName=args[0];
  String title="Java/JOGL GpuElint test application ("+fileName+")";
  new TestGpuElint(title,fileName,800,600);
  }
}

//----------------------------------------------------------------------------
