//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "gpuElint.h"

#if defined _WIN32 //-- Windows-specific implementation ----------------------

static HINSTANCE hInstance;
static HWND window;
static HDC hdc;
static HGLRC wglContext;

#if !defined WGL_CONTEXT_MAJOR_VERSION_ARB
# define WGL_CONTEXT_MAJOR_VERSION_ARB 0x2091
#endif

#if !defined WGL_CONTEXT_MINOR_VERSION_ARB
# define WGL_CONTEXT_MINOR_VERSION_ARB 0x2092
#endif

#if !defined WGL_CONTEXT_LAYER_PLANE_ARB
# define WGL_CONTEXT_LAYER_PLANE_ARB 0x2093
#endif

#if !defined WGL_CONTEXT_FLAGS_ARB
# define WGL_CONTEXT_FLAGS_ARB 0x2094
#endif

#if !defined WGL_CONTEXT_PROFILE_MASK_ARB
# define WGL_CONTEXT_PROFILE_MASK_ARB 0x9126
#endif

#if !defined WGL_CONTEXT_DEBUG_BIT_ARB
# define WGL_CONTEXT_DEBUG_BIT_ARB 0x0001
#endif

#if !defined WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB
# define WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB 0x0002
#endif

#if !defined WGL_CONTEXT_CORE_PROFILE_BIT_ARB
# define WGL_CONTEXT_CORE_PROFILE_BIT_ARB 0x00000001
#endif

#if !defined WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB
# define WGL_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB 0x00000002
#endif

static
HGLRC
(APIENTRY *createContextAttribsFnct)(HDC hDC,
                                     HGLRC,
                                     const int *);
static
LRESULT CALLBACK
wndProc(HWND hWnd,
        UINT uMsg,
        WPARAM wParam,
        LPARAM lParam)
{
// printf("wndProc(hWnd=%d,uMsg=%d,wParam=%d,lParam=%d\n",
//        (int)(intptr_t)hWnd,(int)uMsg,(int)wParam,(int)lParam);
return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

#define WINDOW_CLASS_NAME "GpuElintGlContext"

void
createDefaultGlContext(void)
{
hInstance=(HINSTANCE)GetModuleHandle((LPCTSTR)0);
ON_ERROR(!hInstance,"GetModuleHandle() failure",{ exit(1); });
WNDCLASS wc;
wc.style=CS_OWNDC;
wc.lpfnWndProc=wndProc;
wc.cbClsExtra=0;
wc.cbWndExtra=0;
wc.hInstance=hInstance;
wc.hIcon=LoadIcon((HINSTANCE)0,IDI_APPLICATION);
wc.hCursor=(HCURSOR)0;
wc.hbrBackground=(HBRUSH)0;
wc.lpszMenuName=(LPCTSTR)0;
wc.lpszClassName=WINDOW_CLASS_NAME;
ATOM atm=RegisterClass(&wc);
ON_ERROR(!atm,"RegisterClass() failure",{ exit(1); });
DWORD dwStyle=WS_OVERLAPPEDWINDOW;
DWORD dwExStyle=0;
window=CreateWindowEx(dwExStyle,WINDOW_CLASS_NAME,"",dwStyle,
                      0,0,16,16,(HWND)0,(HMENU)0,hInstance,(LPVOID)0);
ON_ERROR(!window,"CreateWindowEx() failure",{ exit(1); });
PIXELFORMATDESCRIPTOR pfd;
pfd.nSize=sizeof(pfd);
pfd.nVersion=1;
pfd.dwFlags=PFD_DRAW_TO_WINDOW|PFD_SUPPORT_OPENGL|PFD_DOUBLEBUFFER;
pfd.iPixelType=PFD_TYPE_RGBA;
pfd.cColorBits=32;
pfd.cRedBits=0;
pfd.cRedShift=0;
pfd.cGreenBits=0;
pfd.cGreenShift=0;
pfd.cBlueBits=0;
pfd.cBlueShift=0;
pfd.cAlphaBits=0;
pfd.cAlphaShift=0;
pfd.cAccumBits=0;
pfd.cAccumRedBits=0;
pfd.cAccumGreenBits=0;
pfd.cAccumBlueBits=0;
pfd.cAccumAlphaBits=0;
pfd.cDepthBits=24;
pfd.cStencilBits=0;
pfd.cAuxBuffers=0;
pfd.iLayerType=PFD_MAIN_PLANE;
pfd.bReserved=0;
pfd.dwLayerMask=0;
pfd.dwVisibleMask=0;
pfd.dwDamageMask=0;
hdc=GetDC(window);
int pixelFormat=ChoosePixelFormat(hdc,&pfd);
ON_ERROR(!pixelFormat,"ChoosePixelFormat() failure",{ exit(1); });
BOOL r=SetPixelFormat(hdc,pixelFormat,&pfd);
ON_ERROR(!r,"SetPixelFormat() failure",{ exit(1); });
HGLRC tmpContext=wglCreateContext(hdc);
ON_ERROR(!tmpContext,"wglCreateContext() failure",{ exit(1); });
r=wglMakeCurrent(hdc,tmpContext);
ON_ERROR(!r,"wglMakeCurrent() failure",{ exit(1); });
const char *extSuffixes[]={ "", "ARB", "EXT", (const char *)0 };
CoreGl_extensionFunctionWithSuffix(&createContextAttribsFnct,
                                   "wglCreateContextAttribs",
                                   extSuffixes);
if(createContextAttribsFnct)
  {
  int gl3Attribs[]={WGL_CONTEXT_MAJOR_VERSION_ARB,
                    3,
                    WGL_CONTEXT_MINOR_VERSION_ARB,
                    2,
                    WGL_CONTEXT_PROFILE_MASK_ARB,
                    WGL_CONTEXT_CORE_PROFILE_BIT_ARB,
                    // WGL_CONTEXT_FLAGS_ARB,
                    // WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB,
                    0};
  int *useGl3Attribs=(int *)0;
  PFNGLGETSTRINGPROC getStringFnct=(PFNGLGETSTRINGPROC)0;
  CoreGl_extensionFunction(&getStringFnct,"glGetString");
  if(getStringFnct)
    {
    int major=0,minor=0;
    const char *version=(const char *)getStringFnct(GL_VERSION);
    if(version&&(sscanf(version,"%d.%d",&major,&minor)==2)&&
       ((major>3)||((major==3)&&(minor>=2))))
      {
      gl3Attribs[1]=major;
      gl3Attribs[3]=minor;
      useGl3Attribs=gl3Attribs;
      }
    }
  wglContext=createContextAttribsFnct(hdc,(HGLRC)0,useGl3Attribs);
  // ON_ERROR(!wglContext,"wglCreateContextAttribs() failure",{ exit(1); });
  }
if(!wglContext)
  {
  wglContext=tmpContext;
  }
else
  {
  wglMakeCurrent((HDC)0,(HGLRC)0);
  wglDeleteContext(tmpContext);
  r=wglMakeCurrent(hdc,wglContext);
  ON_ERROR(!r,"wglMakeCurrent() failure",{ exit(1); });
  }
}

void
destroyDefaultGlContext(void)
{
wglMakeCurrent((HDC)0,(HGLRC)0);
wglDeleteContext(wglContext);
ReleaseDC(window,hdc);
DestroyWindow(window);
UnregisterClass(WINDOW_CLASS_NAME,hInstance);
}

#else //-- X11-specific implementation ---------------------------------------

#include <X11/Xlib.h>

// FIXME: ugly hack!
// Even in C, the ``class'' member generates a warning!
// --> workaround accessing the ``class'' member right after ``depth''
#define XVisualInfo_class(info) \
        ((&((info)->depth))[1])

static Display *display;
static GLXFBConfig fbConfig;
static XVisualInfo *visual;
static Colormap colormap;
static Window window;
static GLXContext glxContext;

#if !defined GLX_CONTEXT_MAJOR_VERSION_ARB
# define GLX_CONTEXT_MAJOR_VERSION_ARB 0x2091
#endif

#if !defined GLX_CONTEXT_MINOR_VERSION_ARB
# define GLX_CONTEXT_MINOR_VERSION_ARB 0x2092
#endif

#if !defined GLX_CONTEXT_FLAGS_ARB
# define GLX_CONTEXT_FLAGS_ARB 0x2094
#endif

#if !defined GLX_CONTEXT_PROFILE_MASK_ARB
# define GLX_CONTEXT_PROFILE_MASK_ARB 0x9126
#endif

#if !defined GLX_CONTEXT_DEBUG_BIT_ARB
# define GLX_CONTEXT_DEBUG_BIT_ARB 0x0001
#endif

#if !defined GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB
# define GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB 0x0002
#endif

#if !defined GLX_CONTEXT_CORE_PROFILE_BIT_ARB
# define GLX_CONTEXT_CORE_PROFILE_BIT_ARB 0x00000001
#endif

#if !defined GLX_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB
# define GLX_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB 0x00000002
#endif

static
GLXContext
(APIENTRY *createContextAttribsFnct)(Display *,
                                     void *,
                                     GLXContext,
                                     Bool,
                                     const int *);

static
bool // success
checkVisual(Display *display,
            XVisualInfo *info,
            bool doubleBuffer,
            bool stereo,
            bool accum,
            bool stencil)
{
if((info->screen!=DefaultScreen(display))||
   ((XVisualInfo_class(info)!=TrueColor)&&
    (XVisualInfo_class(info)!=DirectColor))||
   (!info->red_mask)||(!info->green_mask)||(!info->blue_mask))
  {
  return false;
  }
int value;
if(glXGetConfig(display,info,GLX_USE_GL,&value)||!value)
  {
  return false;
  }
if(glXGetConfig(display,info,GLX_RGBA,&value)||!value)
  {
  return false;
  }
if(glXGetConfig(display,info,GLX_RED_SIZE,&value)||!value)
  {
  return false;
  }
if(glXGetConfig(display,info,GLX_GREEN_SIZE,&value)||!value)
  {
  return false;
  }
if(glXGetConfig(display,info,GLX_BLUE_SIZE,&value)||!value)
  {
  return false;
  }
#if 0 // sometimes value is 0 even if alpha is supported!
  if(glXGetConfig(display,info,GLX_ALPHA_SIZE,&value)||!value)
    {
    return false;
    }
#endif
if(glXGetConfig(display,info,GLX_DEPTH_SIZE,&value)||!value)
  {
  return false;
  }
if(doubleBuffer)
  {
  if(glXGetConfig(display,info,GLX_DOUBLEBUFFER,&value)||!value)
    {
    return false;
    }
  }
else
  {
  if(glXGetConfig(display,info,GLX_DOUBLEBUFFER,&value)||value)
    {
    return false;
    }
  }
if(stereo)
  {
  if(glXGetConfig(display,info,GLX_STEREO,&value)||!value)
    {
    return false;
    }
  }
else
  {
  if(glXGetConfig(display,info,GLX_STEREO,&value)||value)
    {
    return false;
    }
  }
if(accum)
  {
  if(glXGetConfig(display,info,GLX_ACCUM_RED_SIZE,&value)||!value)
    {
    return false;
    }
  if(glXGetConfig(display,info,GLX_ACCUM_GREEN_SIZE,&value)||!value)
    {
    return false;
    }
  if(glXGetConfig(display,info,GLX_ACCUM_BLUE_SIZE,&value)||!value)
    {
    return false;
    }
  if(glXGetConfig(display,info,GLX_ACCUM_ALPHA_SIZE,&value)||!value)
    {
    return false;
    }
  }
if(stencil)
  {
  if(glXGetConfig(display,info,GLX_STENCIL_SIZE,&value)||!value)
    {
    return false;
    }
  }
return true;
}

static
XVisualInfo *
chooseVisual(Display *display,
             bool doubleBuffer,
             bool stereo,
             bool accum,
             bool stencil)
{
int attribs[0x20];
int attribCount=0;
attribs[attribCount++]=GLX_RGBA;
attribs[attribCount++]=GLX_RED_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_GREEN_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_BLUE_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_ALPHA_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_DEPTH_SIZE;
attribs[attribCount++]=1;
if(doubleBuffer)
  {
  attribs[attribCount++]=GLX_DOUBLEBUFFER;
  }
if(stereo)
  {
  attribs[attribCount++]=GLX_STEREO;
  }
if(accum)
  {
  attribs[attribCount++]=GLX_ACCUM_RED_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_GREEN_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_BLUE_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_ALPHA_SIZE;
  attribs[attribCount++]=1;
  }
if(stencil)
  {
  attribs[attribCount++]=GLX_STENCIL_SIZE;
  attribs[attribCount++]=1;
  }
attribs[attribCount++]=None;
XVisualInfo *best=glXChooseVisual(display,DefaultScreen(display),attribs);
if(best)
  {
  if(!checkVisual(display,best,doubleBuffer,stereo,accum,stencil))
    {
    XFree(best); // reject configurations we don't want to use
    best=(XVisualInfo *)0;
    }
  }
if(!best) // try our own visual selection algorithm
  {
  int infoCount;
  XVisualInfo *infos=XGetVisualInfo(display,VisualNoMask,
                                    (XVisualInfo *)0,&infoCount);
  if(infos)
    {
    for(int i=0;i<infoCount;++i)
      {
      if(checkVisual(display,infos+i,doubleBuffer,stereo,accum,stencil))
        {
        if(!best) // first match
          {
          best=infos+i;
          }
        else if(infos[i].depth>best->depth) // better depth
          {
          best=infos+i;
          }
        else if((infos[i].depth==best->depth)&&           // prefer TrueColor
                (XVisualInfo_class(infos+i)==TrueColor))  // to DirectColor
          {
          best=infos+i;
          }
        }
      }
    if(best)
      {
      XVisualInfo tmp;
      tmp.screen=best->screen;
      tmp.visualid=best->visualid; // get only this visual
      best=XGetVisualInfo(display,VisualScreenMask|VisualIDMask,
                          &tmp,&infoCount);
      }
    XFree(infos);
    }
  }
if(!best&&!doubleBuffer) // try double-buffer if single is not available
  {
  best=chooseVisual(display,true,stereo,accum,stencil);
  }
return best;
}

static
bool // success
checkFBConfig(Display *display,
              GLXFBConfig fbConfig,
              bool doubleBuffer,
              bool stereo,
              bool accum,
              bool stencil)
{
XVisualInfo *visual=glXGetVisualFromFBConfig(display,fbConfig);
if(!visual)
  {
  return false;
  }
if((visual->screen!=DefaultScreen(display))||
   ((XVisualInfo_class(visual)!=TrueColor)&&
    (XVisualInfo_class(visual)!=DirectColor))||
   (!visual->red_mask)||(!visual->green_mask)||(!visual->blue_mask))
  {
  XFree(visual);
  return false;
  }
XFree(visual);
int value;
if(glXGetFBConfigAttrib(display,fbConfig,GLX_RENDER_TYPE,&value)||
   !(value&GLX_RGBA_BIT))
  {
  return false;
  }
#if 0
if(glXGetFBConfigAttrib(display,fbConfig,GLX_SAMPLE_BUFFERS,&value)||
   !value)
  {
  return false;
  }
#endif
if(glXGetFBConfigAttrib(display,fbConfig,GLX_RED_SIZE,&value)||
   !value)
  {
  return false;
  }
if(glXGetFBConfigAttrib(display,fbConfig,GLX_GREEN_SIZE,&value)||
   !value)
  {
  return false;
  }
if(glXGetFBConfigAttrib(display,fbConfig,GLX_BLUE_SIZE,&value)||
   !value)
  {
  return false;
  }
#if 0 // sometimes value is 0 even if alpha is supported!
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ALPHA_SIZE,&value)||
     !value)
    {
    return false;
    }
#endif
if(glXGetFBConfigAttrib(display,fbConfig,GLX_DEPTH_SIZE,&value)||
   !value)
  {
  return false;
  }
if(doubleBuffer)
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_DOUBLEBUFFER,&value)||
     !value)
    {
    return false;
    }
  }
else
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_DOUBLEBUFFER,&value)||
     value)
    {
    return false;
    }
  }
if(stereo)
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_STEREO,&value)||
     !value)
    {
    return false;
    }
  }
else
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_STEREO,&value)||
     value)
    {
    return false;
    }
  }
if(accum)
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ACCUM_RED_SIZE,&value)||
     !value)
    {
    return false;
    }
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ACCUM_GREEN_SIZE,&value)||
     !value)
    {
    return false;
    }
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ACCUM_BLUE_SIZE,&value)||
     !value)
    {
    return false;
    }
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_ACCUM_ALPHA_SIZE,&value)||
     !value)
    {
    return false;
    }
  }
if(stencil)
  {
  if(glXGetFBConfigAttrib(display,fbConfig,GLX_STENCIL_SIZE,&value)||
     !value)
    {
    return false;
    }
  }
return true;
}

static
GLXFBConfig
chooseFBConfig(Display *display,
               bool doubleBuffer,
               bool stereo,
               bool accum,
               bool stencil)
{
int attribs[0x30];
int attribCount=0;
attribs[attribCount++]=GLX_X_RENDERABLE;
attribs[attribCount++]=True;
attribs[attribCount++]=GLX_DRAWABLE_TYPE;
attribs[attribCount++]=GLX_WINDOW_BIT;
attribs[attribCount++]=GLX_RENDER_TYPE;
attribs[attribCount++]=GLX_RGBA_BIT;
attribs[attribCount++]=GLX_X_VISUAL_TYPE;
attribs[attribCount++]=GLX_TRUE_COLOR;
attribs[attribCount++]=GLX_RED_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_GREEN_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_BLUE_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_ALPHA_SIZE;
attribs[attribCount++]=1;
attribs[attribCount++]=GLX_DEPTH_SIZE;
attribs[attribCount++]=1;
if(doubleBuffer)
  {
  attribs[attribCount++]=GLX_DOUBLEBUFFER;
  attribs[attribCount++]=True;
  }
if(stereo)
  {
  attribs[attribCount++]=GLX_STEREO;
  attribs[attribCount++]=True;
  }
if(accum)
  {
  attribs[attribCount++]=GLX_ACCUM_RED_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_GREEN_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_BLUE_SIZE;
  attribs[attribCount++]=1;
  attribs[attribCount++]=GLX_ACCUM_ALPHA_SIZE;
  attribs[attribCount++]=1;
  }
if(stencil)
  {
  attribs[attribCount++]=GLX_STENCIL_SIZE;
  attribs[attribCount++]=1;
  }
// attribs[attribCount++]=GLX_SAMPLE_BUFFERS;
// attribs[attribCount++]=1;
// attribs[attribCount++]=GLX_SAMPLES;
// attribs[attribCount++]=1;
attribs[attribCount++]=None;
GLXFBConfig best=(GLXFBConfig)0;
int configCount;
GLXFBConfig *fbc=glXChooseFBConfig(display,DefaultScreen(display),
                                   attribs,&configCount);
if(fbc)
  {
  int bestSamples=-1;
  for(int i=0;i<configCount;++i)
    {
    if(checkFBConfig(display,fbc[i],doubleBuffer,stereo,accum,stencil))
      {
      int sampleBuffers;
      if(!glXGetFBConfigAttrib(display,fbc[i],GLX_SAMPLE_BUFFERS,
                               &sampleBuffers))
        {
        int samples=0;
        if(!glXGetFBConfigAttrib(display,fbc[i],GLX_SAMPLES,&samples))
          {
          if((!best)||(sampleBuffers&&(samples>bestSamples)))
            {
            best=fbc[i];
            bestSamples=samples;
            }
          }
        }
      }
    }
  XFree(fbc);
  }
if(!best&&!doubleBuffer) // try double-buffer if single is not available
  {
  best=chooseFBConfig(display,true,stereo,accum,stencil);
  }
return best;
}

void
createDefaultGlContext(void)
{
display=XOpenDisplay(getenv("DISPLAY") ? "" : ":0");
ON_ERROR(!display,"XOpenDisplay() failure",{ exit(1); });
const char *extSuffixes[]={ "", "ARB", "EXT", (const char *)0 };
CoreGl_extensionFunctionWithSuffix(&createContextAttribsFnct,
                                   "glXCreateContextAttribs",
                                   extSuffixes);
bool doubleBuffer=true;
bool stereo=false;
bool accum=false;
bool stencil=false;
if(createContextAttribsFnct)
  {
  fbConfig=chooseFBConfig(display,doubleBuffer,stereo,accum,stencil);
  }
if(fbConfig)
  {
  visual=glXGetVisualFromFBConfig(display,fbConfig);
  ON_ERROR(!visual,"glXGetVisualFromFBConfig() failure",{ exit(1); });
  }
if(!visual)
  {
  visual=chooseVisual(display,doubleBuffer,stereo,accum,stencil);
  ON_ERROR(!visual,"chooseVisual() failure",{ exit(1); });
  }
colormap=XCreateColormap(display,DefaultRootWindow(display),
                         visual->visual,AllocNone);
ON_ERROR(!colormap,"XCreateColormap() failure",{ exit(1); });

XSetWindowAttributes wAttr;
wAttr.colormap=colormap;
wAttr.background_pixel=WhitePixel(display,DefaultScreen(display));
wAttr.border_pixel=BlackPixel(display,DefaultScreen(display));
window=XCreateWindow(display,DefaultRootWindow(display),
                     0,0,16,16,0,visual->depth,InputOutput,visual->visual,
                     CWColormap|CWBackPixel|CWBorderPixel,&wAttr);
ON_ERROR(!window,"XCreateWindow() failure",{ exit(1); });
GLXContext tmpContext=glXCreateContext(display,visual,(GLXContext)0,True);
ON_ERROR(!tmpContext,"glXCreateContext() failure",{ exit(1); });
Bool r=glXMakeCurrent(display,window,tmpContext);
ON_ERROR(!r,"glXMakeCurrent() failure",{ exit(1); });
if(fbConfig)
  {
  int gl3Attribs[]={GLX_CONTEXT_MAJOR_VERSION_ARB,
                    3,
                    GLX_CONTEXT_MINOR_VERSION_ARB,
                    2,
                    GLX_CONTEXT_PROFILE_MASK_ARB,
                    GLX_CONTEXT_CORE_PROFILE_BIT_ARB,
                    // GLX_CONTEXT_FLAGS_ARB,
                    // GLX_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB,
                    0};
  int *useGl3Attribs=(int *)0;
  int major=0,minor=0;
  PFNGLGETSTRINGPROC getStringFnct=(PFNGLGETSTRINGPROC)0;
  CoreGl_extensionFunction(&getStringFnct,"glGetString");
  if(getStringFnct)
    {
    const char *version=(const char *)getStringFnct(GL_VERSION);
    if(version&&(sscanf(version,"%d.%d",&major,&minor)==2)&&
       ((major>3)||((major==3)&&(minor>=2))))
      {
      gl3Attribs[1]=major;
      gl3Attribs[3]=minor;
      useGl3Attribs=gl3Attribs;
      }
    }
  glxContext=createContextAttribsFnct(display,fbConfig,(GLXContext)0,
                                      True,useGl3Attribs);
  // ON_ERROR(!glxContext,"glXCreateContextAttribs() failure",{ exit(1); });
  }
if(!glxContext)
  {
  glxContext=tmpContext;
  }
else
  {
  glXMakeCurrent(display,(Window)0,(GLXContext)0);
  glXDestroyContext(display,tmpContext);
  r=glXMakeCurrent(display,window,glxContext);
  ON_ERROR(!r,"glXMakeCurrent() failure",{ exit(1); });
  }
XSync(display,False);
}

void
destroyDefaultGlContext(void)
{
XSync(display,False);
glXMakeCurrent(display,(Window)0,(GLXContext)0);
glXDestroyContext(display,glxContext);
XDestroyWindow(display,window);
XFreeColormap(display,colormap);
XFree(visual);
XCloseDisplay(display);
}

#endif

//----------------------------------------------------------------------------
