//----------------------------------------------------------------------------
// Copyright (c) 2015-2016 Fabrice HARROUET (ENIB)
//
// This file is part of GpuElint.
//   http://www.enib.fr/~harrouet/gpuelint.html
//
// GpuElint is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// GpuElint is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with GpuElint.  If not, see <http://www.gnu.org/licenses/>.
//----------------------------------------------------------------------------

#include "common.h"

typedef int (*AnalyseFnct)(GpuElint *,const char *,int,int *,Sequence *);

void
analysePulseFile(GpuElint *ge,
                 const char *fileName)
{
int pulseCount=readPulseCount(fileName);
int *marks=Memory_alloc(pulseCount,int);
AnalyseFnct fncts[16];
int fnctCount=0;
if(USE_ID_SERIAL)   { fncts[fnctCount++]=ID_serialAnalysis; }
if(USE_CP_SERIAL)   { fncts[fnctCount++]=CP_serialAnalysis; }
#if defined __cplusplus
if(USE_ID_PARALLEL) { fncts[fnctCount++]=ID_parallelAnalysis; }
if(USE_CP_PARALLEL) { fncts[fnctCount++]=CP_parallelAnalysis; }
#endif
if(USE_ID_GPU)      { fncts[fnctCount++]=ID_gpuAnalysis; }
if(USE_BM_GPU)      { fncts[fnctCount++]=BM_gpuAnalysis; }
Sequence sequences[MAX_SEQUENCE_COUNT];
for(int f=0;f<fnctCount;++f)
  {
  printf("#######################################"
         "#######################################\n");
  int sequenceCount=(*fncts[f])(ge,fileName,pulseCount,marks,sequences);
  int remaining=pulseCount;
  for(int s=0;s<sequenceCount;++s)
    {
    if(s<MAX_SEQUENCE_COUNT)
      {
      const int count=sequences[s].pulseCount;
      remaining-=count;
      printf("Sequence [%3d] %9d pulses, %11.3f us (%11.3f Hz)\n",
             s,count,1e6f*sequences[s].pri,1.0f/sequences[s].pri);
      }
    else
      {
      printf("... %d more sequences\n",
             sequenceCount-MAX_SEQUENCE_COUNT);
      break;
      }
    }
  printf("%d remaining pulses\n",remaining);
  }
printf("#######################################"
       "#######################################\n");
Memory_free(marks);
}

int
main(int argc,
     char **argv)
{
if(argc!=2)
  { fprintf(stderr,"usage: %s pulse_file.raw\n",argv[0]); exit(1); }
createDefaultGlContext();
GpuElint *ge=GpuElint_new((GpuElint *)0);
if(!ge)
  {
  fprintf(stderr,"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  fprintf(stderr,"~~~~ GpuElint initialisation failure ~~~~\n");
  fprintf(stderr,"~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  exit(1);
  }
const GpuElintProperties *properties=GpuElint_properties(ge);
printf("DTOA_SKIP_COUNT=%d\n",
       properties->DTOA_SKIP_COUNT);
printf("MAX_MATCH_COUNT=%d\n",
       properties->MAX_MATCH_COUNT);
printf("glVersionMajor=%d\n",
       properties->glVersionMajor);
printf("glVersionMinor=%d\n",
       properties->glVersionMinor);
printf("glslVersion=%d\n",
       properties->glslVersion);
printf("maxWorkGroupCount=%d %d %d\n",
       properties->maxWorkGroupCount[0],
       properties->maxWorkGroupCount[1],
       properties->maxWorkGroupCount[2]);
printf("maxWorkGroupSize=%d %d %d\n",
       properties->maxWorkGroupSize[0],
       properties->maxWorkGroupSize[1],
       properties->maxWorkGroupSize[2]);
printf("maxWorkGroupInvocations=%d\n",
       properties->maxWorkGroupInvocations);
printf("maxUniformBlocks=%d\n",
       properties->maxUniformBlocks);
printf("maxTextureImageUnits=%d\n",
       properties->maxTextureImageUnits);
printf("maxAtomicCounterBuffers=%d\n",
       properties->maxAtomicCounterBuffers);
printf("maxAtomicCounters=%d\n",
       properties->maxAtomicCounters);
printf("maxSharedMemorySize=%d\n",
       properties->maxSharedMemorySize);
printf("maxUniformComponents=%d\n",
       properties->maxUniformComponents);
printf("maxImageUniforms=%d\n",
       properties->maxImageUniforms);
printf("maxCombinedUniformComponents=%d\n",
       properties->maxCombinedUniformComponents);
printf("maxShaderStorageBlocks=%d\n",
       properties->maxShaderStorageBlocks);
printf("warpSize=%d\n",
       properties->warpSize);
printf("warpsPerStreamingMultiprocessor=%d\n",
       properties->warpsPerStreamingMultiprocessor);
printf("streamingMultiprocessorCount=%d\n",
       properties->streamingMultiprocessorCount);
printf("hasBallot=%d\n",
       properties->hasBallot);
printf("hasShuffle=%d\n",
       properties->hasShuffle);
printf("OpenGL version: %s\n",
       glGetString(GL_VERSION));
printf("          GLSL: %s\n",
       glGetString(GL_SHADING_LANGUAGE_VERSION));
printf("        vendor: %s\n",
       glGetString(GL_VENDOR));
printf("      renderer: %s\n",
       glGetString(GL_RENDERER));
analysePulseFile(ge,argv[1]);
GpuElint_delete(ge);
destroyDefaultGlContext();
return 0;
}

//----------------------------------------------------------------------------
